"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Main class for turbine modelling functions
"""

import numpy as np
from zutil.zWind import ForceModels, zController, zBem


class zTurbine:
    def __init__(self, control_dict: dict) -> None:
        """Base class for turbine and propellor modelling"""
        self.name = control_dict["name"]

        # Geometry properties
        self.centre = control_dict["geometry"]["centre"]
        self.up = control_dict["geometry"]["up"]
        self.normal = control_dict["geometry"]["normal"]
        self.inner_radius = control_dict["geometry"]["inner radius"]
        self.outer_radius = control_dict["geometry"]["outer radius"]
        self.reference_plane = control_dict["reference plane"]

        # self.refrence_point = control_dict["reference point"]
        self.update_frequency = control_dict["update frequency"]

        # Model properties
        self.rotation_direction = control_dict.get("rotation direction", "clockwise")

        # Store remaining dictionaries incase they're needed later
        self.discretisation = control_dict["discretisation"]
        self.model = control_dict["model"]
        self.geometry = control_dict["geometry"]
        self.control = control_dict["controller"]
        self.spec = control_dict

        self.rotor_swept_area = np.pi * (
            self.outer_radius**2 - self.inner_radius**2
        )

        self.verbose = control_dict.get("verbose", False)

        self.assign_model()
        self.assign_controller()

    def assign_model(self) -> None:
        """Assign the physics model to be applied when calculating the force on the turbine

        The turbine can either be modelled using an aggregated thrust and torque curve, or modelled as individual blade elements
        """
        self.model_type = self.model["type"]

        # switch based on model type to load model
        if self.model_type == "simple":
            power_model = self.model["power model"]
            if power_model == "fixed":
                kind = self.model["kind"]
                if kind == "turbine":
                    self.turbine_model = ForceModels.zTurbineSimpleFixed(self.model)
                elif kind == "propellor":
                    self.turbine_model = ForceModels.zPropellorSimpleFixed(self.model)
            elif power_model == "curve":
                self.turbine_model = ForceModels.zTurbineSimpleCurve(self.model)

        elif self.model_type == "induction":
            power_model = self.model["power model"]
            if power_model == "fixed":
                self.turbine_model = ForceModels.zTurbineInductionFixed(self.model)
            elif power_model == "curve":
                self.turbine_model = ForceModels.zTurbineIndutionCurve(self.model)

        # BEMT options
        elif self.model_type == "BET":
            self.turbine_model = zBem.zTurbineBetFull(self.model)
        elif self.model_type == "BEMT":
            self.turbine_model = zBem.BEMT_Turbine(self.model)

        if self.verbose:
            print("Loaded: {}".format(self.turbine_model.model_type))

    def assign_controller(self) -> None:
        self.controller_type = self.control["type"]

        # switch based on controller type to load controller
        if self.controller_type == "fixed":
            self.controller = zController.FixedController(self.control)
        elif self.controller_type == "tsr curve":
            self.controller = zController.TSRController(self.control)
        elif self.controller_type == "schedule":
            self.controller = zController.zControllerSchedule(self.control)
        else:
            self.controller = zController.ControllerBase(self.control)

        if self.verbose:
            print("Loaded: {}".format(self.controller.controller_type))

    def get_performance(self, u_ref: float) -> None:
        """Get the total power and performance of the whole turbine"""
        if self.controller.iterative:
            pass
        else:
            self.pitch, self.omega = self.controller.set_conditions(u_ref, self)

        thrust, power, torque = self.turbine.get_performance(u_ref, self, 1.225)

        self.discretisation.discretise_performance(thrust, torque)

    def get_discrete_performance(self, u_ref: float, seg: object) -> None:
        """get the performance of a discretised section of a turbine disc- could be driven by actuator disc or line model"""

        # controller step- figure out operating conditions of turbine for given inflow conditions
        if self.controller.iterative:
            pass
        else:
            self.pitch, self.omega = self.controller.set_conditions(u_ref, self)

        # simulation step- use the force model to predict the power on the models
        dt, dq = self.turbine_model.discrete_performance(u_ref, 1.225, seg, self)
        return dt, dq



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Force models used in zWind
"""

import numpy as np
from zutil.zWind import zTurbine
from zutil.converters import radps2rps


class zTurbineBase:
    def __init__(self, model_dict: dict) -> None:
        """Base class for turbine and propellor modelling"""
        pass


class zSimpleFixed(zTurbineBase):
    def __init__(self, model_dict: dict) -> None:
        super().__init__(model_dict)

        self.power = None
        self.power_coefficient = None
        self.thrust = None
        self.thrust_coefficient = None

    def get_power(self, u_ref: float) -> float:
        return self.power

    def get_power_coefficient(self, u_ref: float) -> float:
        return self.power_coefficient

    def get_thrust(self, u_ref: float) -> float:
        return self.thrust

    def get_thrust_coefficient(self, u_ref: float) -> float:
        return self.thrust_coefficient

    def get_performance(
        u_ref: float, density: float, turbine: zTurbine
    ) -> tuple[float, float, float]:
        return (0.0, 0.0, 0.0)

    def discrete_performance(
        self, u_ref: float, density: float, seg: object, turbine: zTurbine
    ) -> tuple:
        thrust, power, torque = self.get_performance(u_ref, density, turbine)

        dt = thrust * (seg.da / turbine.rotor_swept_area)
        dq = torque * (seg.da / (turbine.rotor_swept_area * seg.rp))

        if turbine.rotation_direction == "anticlockwise":
            dq = -dq

        return (dt, dq)


class zTurbineSimpleFixed(zSimpleFixed):
    """Turbine with simple power model at fixed operating point- no controller, no curve"""

    def __init__(self, model_dict: dict) -> None:
        super().__init__(model_dict)
        self.model_type = "Turbine Simple Fixed"

        self.power = model_dict["power"]
        self.thrust_coefficient = model_dict["thrust coefficient"]

    def get_performance(self, u_ref: float, density: float, turbine: zTurbine) -> tuple:
        thrust_coefficient = self.get_thrust_coefficient(u_ref)
        power = self.get_power(u_ref)

        pitch, omega = turbine.controller.get_conditions()

        if omega != 0.0:
            torque = power / omega
            thrust = (
                0.5
                * density
                * u_ref**2
                * turbine.rotor_swept_area
                * thrust_coefficient
            )
        else:
            torque = 0.0
            thrust = 0.0
            power = 0.0

        return (thrust, power, torque)


class zPropellorSimpleFixed(zSimpleFixed):
    """Propellor with simple power model at fixed operating point- no controller, no curve"""

    def __init__(self, model_dict: dict) -> None:
        super().__init__(model_dict)
        self.model_type = "Propellor Simple Fixed"

        self.thrust_coefficient = model_dict["thrust coefficient"]
        self.power_coefficient = model_dict["power coefficient"]

    def get_performance(self, u_ref: float, density: float, turbine: zTurbine) -> tuple:
        ct = self.get_thrust_coefficient(u_ref)
        cp = self.get_power_coefficient(u_ref)

        pitch, omega = turbine.controller.get_conditions()

        if omega != 0.0:
            power = (
                cp * density * radps2rps(omega) ** 3 * (turbine.outer_radius * 2) ** 5
            )
            torque = power / omega
            thrust = ct * radps2rps(omega) ** 2 * (turbine.outer_radius * 2) ** 4
        else:
            thrust = 0.0
            power = 0.0
            torque = 0.0

        return (thrust, power, torque)


class zTurbineSimpleCurve(zTurbineSimpleFixed):
    """Turbine with simple power model with power curve"""

    def __init__(self, model_dict: dict) -> None:
        super().__init__(model_dict)
        self.model_type = "Turbine Simple Curve"

        self.cp_curve = np.array(model_dict["power curve"])
        self.ct_curve = np.array(model_dict["thrust coefficient curve"])

    def get_power(self, u_ref: float) -> float:
        return np.interp(
            u_ref, self.cp_curve[:, 0], self.cp_curve[:, 1], left=0, right=0
        )

    def get_thrust_coefficient(self, u_ref: float) -> float:
        return np.interp(
            u_ref, self.ct_curve[:, 0], self.ct_curve[:, 1], left=0, right=0
        )


class zTurbineInductionFixed(zTurbineBase):
    """"""

    def __init__(self, control_dict: dict) -> None:
        super().__init__(control_dict)
        self.model_type = "Turbine induction"

    def get_axial_induction_factor(self, ct: float) -> float:
        return (4.0 - np.sqrt(4.0 * 4.0 - 4.0 * 4.0 * ct)) / (2.0 * 4.0)

    def discrete_performance(self, u_ref: float, density: float, seg: object) -> tuple:
        ct = self.get_thrust_coefficient(u_ref)
        tsr = self.get_tsr(u_ref)

        ind_fac = self.get_axial_induction_factor(ct)

        omega = (tsr * u_ref) / self.outer_radius

        dt = 0.5 * density * u_ref**2 * seg.da * 4.0 * ind_fac * (1.0 - ind_fac)

        lambda_r = (seg.rp * omega) / u_ref
        if lambda_r > 0.0:
            angular_induction_factor = -0.5 + np.sqrt(
                0.25 + ((ind_fac * (1.0 - ind_fac)) / (lambda_r**2))
            )
        else:
            angular_induction_factor = 0.0

        dq = (
            4.0
            * angular_induction_factor
            * (1 - ind_fac)
            * 0.5
            * density
            * u_ref
            * omega
            * seg.rp
            * seg.rp
            * seg.da
            * seg.rp
        )

        if self.rotation_direction == "anticlockwise":
            dq = -dq

        return (dt, dq)



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Init module for zutil.post, requires paraview.simple
"""

import os

try:
    import paraview as pv
    import paraview.simple as pvs

    from .post import get_case_parameters
    from .post import get_case_parameters_str
    from .post import get_case_success

    from .post import get_case_root, get_case_report
    from .post import for_each

    from .post import ProgressBar
    from .post import get_status_dict

    from .post import cp_profile
    from .post import cf_profile

    from .post import calc_lift_centre_of_action
    from .post import calc_drag_centre_of_action
    from .post import get_monitor_data
    from .post import calc_moment

    from .post import calc_force_wall
    from .post import calc_moment_wall
    from .post import calc_force

    from .post import cp_profile_wall_from_file
    from .post import cf_profile_wall_from_file

    from .post import get_num_procs

    from .post import clean_vtk

    from .post import vtk_logo_stamp
    from .post import vtk_text_stamp

    pv.compatibility.major = 5
    pv.compatibility.minor = 4
    pvs._DisableFirstRenderCameraReset()
except ImportError:
    # check if in zCFD environment
    if "ZCFD_HOME" in os.environ:
        raise ImportError(
            "paraview.simple not found in zCFD environment. "
            "Please check your installation."
        )
    else:
        raise ImportError(
            "ERROR: zutil.post requires paraview.simple to be installed- please use the zCFD environment to run this module."
        ) from None



"""
Copyright (c) 2012-2025, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

acoustic post processing functions
"""

import numpy as np
from collections.abc import Iterable
from typing import Union
from scipy import signal


def calculate_PSD(
    p: list,
    time: Union[Iterable, float],
    sampling_frequency: float = 1.0,
) -> tuple:
    """
    Calculate Power Spectral Density (PSD) using Welch's method.

    Args:
        T (Arraylike): Time data array
        p (Arraylike): Pressure data array
        sampling_frequency (float = 1.0): Sampling frequency in Hz.
    Returns:
        (frequency, psd values): Frequencies and corresponding PSD values.
    """
    if isinstance(time, Iterable):
        # If time is a list, calculate the time step (dT)
        dT = time[1] - time[0]
    elif isinstance(time, float):
        # If time is a float, use it as the time step (dT)
        dT = time
    else:
        raise ValueError("Time must be a list or a float")

    # Parameters for Welch's method in frequency analysis
    fs = 1.0 / dT  # Sampling frequency
    nps = (1.0 / sampling_frequency) * fs  # Number of points per segment

    # Compute Power Spectral Density (PSD) using Welch's method
    f, Pxx_den = signal.welch(p, fs, nperseg=nps, scaling="density")  # Welch's method

    return f, Pxx_den


def convert_to_dB(Pxx_den: list, inv_pref: float = 1.0 / 2e-5) -> list:
    """
    Convert Power Spectral Density (PSD) to decibels (dB).

    Args:
        Pxx_den (Arraylike): Power Spectral Density values.
        inv_pref (float = 1.0/2e-5): Inverse reference pressure for SPL in dB.
    Returns:
        Arraylike: Converted PSD values in dB.
    """
    return [10 * np.log10(xx * inv_pref**2) for xx in Pxx_den]


def calculate_thirdoctave_bands(
    f: list,
    Pxx_den: list,
) -> tuple[list, list]:
    """
    Calculate third-octave bands from Power Spectral Density (PSD).

    Args:
        f (Arraylike): Frequency data array.
        Pxx_den (Arraylike): Power Spectral Density values.
    Returns:
        (band_frequencies, band_power): Third-octave band data.
    """
    # Define third-octave frequency bands
    third_octave_bands = [
        [11.2, 12.5, 14.1],
        [14.1, 16, 17.8],
        [17.8, 20, 22.4],
        [22.4, 25, 28.2],
        [28.2, 31.5, 35.5],
        [35.5, 40, 44.7],
        [44.7, 50, 56.2],
        [56.2, 63, 70.8],
        [70.8, 80, 89.1],
        [89.1, 100, 112],
        [112, 125, 141],
        [141, 160, 178],
        [178, 200, 224],
        [224, 250, 282],
        [282, 315, 355],
        [355, 400, 447],
        [447, 500, 562],
        [562, 630, 708],
        [708, 800, 891],
        [891, 1000, 1122],
        [1122, 1250, 1413],
        [1413, 1600, 1778],
        [1778, 2000, 2239],
        [2239, 2500, 2818],
        [2818, 3150, 3548],
        [3548, 4000, 4467],
        [4467, 5000, 5623],
        [5623, 6300, 7079],
        [7079, 8000, 8913],
        [8913, 10000, 11220],
        [11220, 12500, 14130],
        [14130, 16000, 17780],
        [17780, 20000, 22390],
    ]

    # Initialize data structures for third-octave band processing
    third_octave_data = [0.0 for i in third_octave_bands]
    third_octave_freq = [band[1] for band in third_octave_bands]
    df = f[1] - f[0]

    # Populate third-octave band data by summing within each band's frequency range
    for freq, psd in zip(f, Pxx_den):
        bin_min = freq - 0.5 * df
        bin_max = freq + 0.5 * df
        if 11.2 <= freq <= 22390.0:
            for j, band in enumerate(third_octave_bands):
                band_min, band_center, band_max = band
                if bin_min >= band_min and bin_max < band_max:
                    third_octave_data[j] += psd * df
                elif bin_min < band_min and bin_max >= band_max:
                    third_octave_data[j] += psd * (band_max - band_min)
                elif bin_min < band_min and bin_max >= band_min:
                    third_octave_data[j] += psd * (bin_max - band_min)
                elif bin_min < band_max and bin_max >= band_max:
                    third_octave_data[j] += psd * (band_max - bin_min)

    # clip values to avoid undefined behaviour
    non_zero_indices = np.nonzero(third_octave_data)
    if len(non_zero_indices[0]) == 0:
        raise ValueError("No non-zero elements found in third-octave data")
    else:
        start, end = non_zero_indices[0][0], non_zero_indices[0][-1] + 1
        filtered_freq = third_octave_freq[start:end]
        filtered_data = third_octave_data[start:end]

    return filtered_freq, filtered_data



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Helper functions for accessing zCFD Paraview functionality
"""

from tqdm import tqdm
from zutil import analysis
import json

import paraview.simple as pvs
from builtins import object
from builtins import range
from builtins import str
from future import standard_library
from typing import Union, Tuple, Optional
import os
from zutil.fileutils import clean_name
from zutil.fileutils import get_csv_data
import pandas as pd
from pathlib import Path
import matplotlib.image as image
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from zutil.fileutils import _get_logo_path

standard_library.install_aliases()
# from paraview.vtk.util import numpy_support
try:
    from paraview.vtk.dataset_adapter import Table
    from paraview.vtk.dataset_adapter import DataSet
    from paraview.vtk.dataset_adapter import PointSet
except ImportError:
    from paraview.vtk.numpy_interface.dataset_adapter import Table
    from paraview.vtk.numpy_interface.dataset_adapter import DataSet
    from paraview.vtk.numpy_interface.dataset_adapter import PointSet


def get_data_array_and_zone(input, array_name, cell_data):
    """"""
    if cell_data:
        p = input.GetCellData().GetArray(array_name)
        z = input.GetCellData().GetArray("zone")
        n = input.GetNumberOfCells()
    else:
        # check if requested data array excists in point data
        p = input.GetPointData().GetArray(array_name)
        z = input.GetPointData().GetArray("zone")
        n = input.GetNumberOfPoints()

    return p, z, n


def _call_extract(source, proxy, array_name, ignore_zone, filter):
    if array_name in source.CellData.keys():
        extract = sum_and_zone_filter(
            proxy, array_name, ignore_zone, filter, cell_data=True
        )
    elif array_name in source.PointData.keys():
        extract = sum_and_zone_filter(
            proxy, array_name, ignore_zone, filter, cell_data=False
        )
    else:
        print("Error: {} variable not found in input".format(array_name))
        extract = [0.0, 0.0, 0.0]
    return extract


def sum_and_zone_filter_array(
    input: any, array_name: str, ignore_zone: list, filter: any = None, cell_data=True
) -> list:
    """Take an array data set and sum the values in the valid regions as specified through the filters"""

    sum = [0.0, 0.0, 0.0]
    # check if requested data array exists in cell data:
    p, z, num_locations = get_data_array_and_zone(input, array_name, cell_data)

    for x in range(num_locations):
        if len(ignore_zone) == 0:
            v = p.GetTuple(x)
            for i in range(0, 3):
                sum[i] += v[i]
        else:
            zone = z.GetValue(x)
            if zone not in ignore_zone:
                v = p.GetTuple(x)
                if filter is None or filter.test(input, x):
                    # print 'Zone: %i'%(zone)
                    for i in range(0, 3):
                        sum[i] += v[i]
    return sum


def sum_and_zone_filter(
    input: any, array_name: str, ignore_zone: list, filter: any = None, cell_data=True
) -> list:
    """Break down vtkMultiBlockDataSet object and sum desired arrays according to filters"""
    sum = [0.0, 0.0, 0.0]
    if input.IsA("vtkMultiBlockDataSet"):
        iter = input.NewIterator()
        iter.UnRegister(None)
        iter.InitTraversal()
        while not iter.IsDoneWithTraversal():
            cur_input = iter.GetCurrentDataObject()
            v = sum_and_zone_filter_array(
                cur_input, array_name, ignore_zone, filter, cell_data
            )
            for i in range(0, 3):
                sum[i] += v[i]
            iter.GoToNextItem()
    else:
        sum = sum_and_zone_filter_array(
            input, array_name, ignore_zone, filter, cell_data
        )

    return sum


class GeomFilterLT(object):
    '''Geometric filter, which tests whether an input (x,y,z)[idx] is less than a float "val"'''

    def __init__(self, val: int, idx: int) -> None:
        self.val = val
        self.idx = idx

    def test(self, input: any, x: int) -> bool:
        """Returns True if the (x,y,z)[idx] value is less than the filter threshold value"""
        centre = input.GetCellData().GetArray("centre").GetTuple(x)
        if centre[self.idx] < self.val:
            return True
        else:
            return False


class GeomFilterGT(object):
    '''Geometric filter, which tests whether an input (x,y,z)[idx] is greater than a float "val"'''

    def __init__(self, val: float, idx: int) -> None:
        #
        self.val = val
        self.idx = idx

    def test(self, input: any, x: int) -> bool:
        """Returns True if the (x,y,z)[idx] value is greater than the filter threshold value"""
        centre = input.GetCellData().GetArray("centre").GetTuple(x)
        if centre[self.idx] >= self.val:
            return True
        else:
            return False


def clean_vtk(
    vtk_object: object, cellDataToPointData: bool = True, mergeBlocks: bool = True
):
    """Takes a VTK object, and performs basic cleanup operations:
    cleanToGrid
    mergeBlocks [optional]
    cellDataToPointData [optional]

    Args:
        vtk_object: A loaded vtk object
    Optional:
        mergeBlocks: default=True-Whether to perform a mergeBlocks operation on the data
        cellDataToPointData: defult=True- Whether to perform a cellDataToPointData mapping on the object
    """
    data = pvs.CleantoGrid(Input=vtk_object)

    if mergeBlocks:
        data = pvs.MergeBlocks(Input=data)
        pvs.UpdatePipeline()

    if cellDataToPointData:
        data = pvs.CellDatatoPointData(Input=data)
        data.ProcessAllArrays = 1
        data.PassCellData = 1
        pvs.UpdatePipeline()

    return data


def calc_force_from_file(
    file_name: str,
    ignore_zone: list,
    half_model: bool = False,
    filter: any = None,
    **kwargs,
) -> Tuple[float, float]:
    """Calculate the pressure and friction force

    This function requires that the VTK file contains three cell data arrays
    called pressureforce, frictionforce and zone

    Args:
        file_name (str): the VTK file name including path
        ignore_zone (list): List of zones to be ignored

    Kwargs:
        half_nodel (bool): Does the data represent only half of the model
        filter (function): GeomFilter object

    Returns:
        float, float. pressure force and friction force
    """
    wall = pvs.PVDReader(FileName=file_name)
    wall.UpdatePipeline()

    return calc_force(wall, ignore_zone, half_model, filter, kwargs)


def calc_force_wall(
    file_root: str,
    ignore_zone: list,
    half_model: bool = False,
    filter: any = None,
    **kwargs,
) -> Tuple[float, float]:
    """Calculate pressure and friction forces at wall"""

    wall = pvs.PVDReader(FileName=file_root + "_wall.pvd")
    wall.UpdatePipeline()

    force = calc_force(wall, ignore_zone, half_model, filter, **kwargs)
    pvs.Delete(wall)
    del wall
    return force


def vtk_logo_stamp(
    input: any, location="Upper Left Corner", logo_file=None, strapline=False
):
    """stamps a render view with a zcfd logo
    Inputs:
    input: paraview render view object
    Optional:
    location: str (default="Upper Left Corner") location of the logo
    logo_file: str (default=None) optional string to alternative logo file
    strapline: bool (default=False) whether to include the strapline in the logo"""
    logo1 = pvs.Logo()

    # get the logo file
    if logo_file:
        logo_path = logo_file
    else:
        logo_path = _get_logo_path(strapline)

    logo_texture = pvs.CreateTexture(logo_path)
    logo1.Texture = logo_texture
    logoDisplay = pvs.Show(logo1, input)
    logoDisplay.WindowLocation = location
    # TODO: Add the ability to adjust the logo size with paraview 13

    return logoDisplay


def vtk_text_stamp(
    input: any,
    text_str: str,
    fontsize=40,
    color=[1.0, 1.0, 1.0],
    location="Lower Left Corner",
    bold=False,
    justification=None,
    font="Courier",
):
    """stamps a renderview object with a text string
    Inputs:
    input: paraview render view object
    text_str: str text to be displayed
    Optional:
    fontsize: int (default=40) font size
    color: list (default=[1.0,1.0,1.0]) RGB color
    location: str (default="Lower Left Corner") location of the text
    bold: bool (default=False) whether the text is bold
    justification: str (default=None) justification of the text- by default it is set by the location
    font: str (default="Courier") font family
    """
    text = pvs.Text()
    text.Text = text_str
    textDisplay = pvs.Show(text, input)
    textDisplay.FontSize = fontsize
    textDisplay.Color = color
    textDisplay.Bold = int(bold)
    textDisplay.WindowLocation = location
    textDisplay.FontFamily = font
    if justification:
        textDisplay.Justification = justification
    else:
        textDisplay.Justification = location.split(" ")[1]
    return textDisplay


def calc_force(
    surface_data: any,
    ignore_zone: list = [],
    half_model: bool = False,
    filter: any = None,
    **kwargs,
) -> Tuple[float, float]:
    """Calculate forces from a surface file

    Inputs:
    surface_data: VTKPointData object

    Optional:
    ignore_zones: list (default = [])
    half_model: bool (default = False)
    filter: callable (default = None)
    """

    sum_client = pvs.servermanager.Fetch(surface_data)

    pforce = _call_extract(
        surface_data, sum_client, "pressureforce", ignore_zone, filter
    )
    fforce = _call_extract(
        surface_data, sum_client, "frictionforce", ignore_zone, filter
    )

    if half_model:
        for i in range(0, 3):
            pforce[i] *= 2.0
            fforce[i] *= 2.0

    del sum_client

    return pforce, fforce


def calc_moment_wall(
    file_root: str,
    ignore_zone: list,
    half_model: bool = False,
    filter: any = None,
    **kwargs,
) -> Tuple[float, float]:
    """Calculate the pressure and friction moment

    This function requires that the _wall.pvd file contains three cell data arrays
    called pressuremomentx, frictionmomentx and zone

    Args:
        file_name (str): the VTK file name including path
        ignore_zone (list): List of zones to be ignored

    Kwargs:
        half_nodel (bool): Does the data represent only half of the model
        filter (function): GeomFilter object

    Returns:
        float, float. pressure force and friction force
    """
    wall = pvs.PVDReader(FileName=file_root + "_wall.pvd")
    wall.UpdatePipeline()

    moment = calc_moment(wall, ignore_zone, half_model, filter, **kwargs)
    pvs.Delete(wall)
    del wall
    return moment


def calc_moment(
    surface_data: any,
    ignore_zone: list = [],
    half_model: bool = False,
    filter: any = None,
    **kwargs,
) -> Tuple[float, float]:
    """Akin to zutil.post.calc_force: Function called by .calc_moment_wall"""

    if "ref_pt" in kwargs:
        p_metric = "pressuremomentx"
        f_metric = "frictionmomentx"
    else:
        p_metric = "pressuremoment"
        f_metric = "frictionmoment"

    sum_client = pvs.servermanager.Fetch(surface_data)

    pmoment = _call_extract(surface_data, sum_client, p_metric, ignore_zone, filter)
    fmoment = _call_extract(surface_data, sum_client, f_metric, ignore_zone, filter)

    if half_model:
        # This is only valid for X-Z plane reflection
        pmoment[0] += -pmoment[0]
        pmoment[1] += pmoment[1]
        pmoment[2] += -pmoment[2]

        fmoment[0] += -fmoment[0]
        fmoment[1] += fmoment[1]
        fmoment[2] += -fmoment[2]

    return pmoment, fmoment


def calc_lift_centre_of_action(
    force: Union[list, tuple], moment: Union[list, tuple], ref_point: Union[list, tuple]
) -> Tuple[tuple, float]:
    """Calculate centre of action of lift given a force, moement and reference point"""
    # longitudinal centre xs0 at zs0
    # spanwise centre ys0 at zs0
    # residual Mz moment (Mx=My=0) mzs0

    xs0 = ref_point[0] - moment[1] / force[2]
    ys0 = ref_point[1] + moment[0] / force[2]

    zs0 = ref_point[2]
    mzs0 = moment[2] - force[1] * (xs0 - ref_point[0]) + force[0] * (ys0 - ref_point[1])

    return (xs0, ys0, zs0), mzs0


def calc_drag_centre_of_action(
    force: Union[list, tuple], moment: Union[list, tuple], ref_point: Union[list, tuple]
) -> Tuple[tuple, float]:
    """Calculate centre of action of drag given a force, moement and reference point"""

    # longitudinal centre xs0 at zs0
    # spanwise centre ys0 at zs0
    # residual Mz moment (Mx=My=0) mzs0

    zs0 = ref_point[2] + moment[1] / force[0]
    ys0 = ref_point[1] - moment[2] / force[0]

    xs0 = ref_point[0]
    # moment[2] - force[1]*(xs0-ref_point[0]) + force[0]*(ys0-ref_point[1])
    mzs0 = 0.0

    return (xs0, ys0, zs0), mzs0


def get_span(wall: any) -> Tuple[float, float]:
    """Returns the min and max y ordinate

    Args:
        wall (vtkMultiBlockDataSet): The input surface

    Returns:
        (float,float). Min y, Max y
    """
    bounds = wall.GetDataInformation().GetBounds()
    y_min = bounds[2]
    y_max = bounds[3]

    return y_min, y_max


def get_chord(
    slice: any,
) -> Tuple[float, float]:
    """Returns the min and max x ordinate

    Args:
        wall (vtkMultiBlockDataSet): The input surface

    Returns:
        (float,float). Min x, Max x
    """
    bounds = slice.GetDataInformation().GetBounds()
    x_min = bounds[0]
    x_max = bounds[1]

    return x_min, x_max


def get_monitor_data(file: str, monitor_name: str, var_name: str) -> tuple:
    """Return the _report file data corresponding to a monitor point and variable name"""
    df = get_csv_data(file, True)
    names = list(df.keys())
    variable = str(monitor_name) + "_" + str(var_name)
    if variable in names:
        return (df[names[0]].tolist(), df[variable].tolist())
    else:
        print(
            "POST.PY: MONITOR POINT: "
            + str(monitor_name)
            + "_"
            + str(var_name)
            + " NOT FOUND in "
            + str(names)
        )


def for_each(surface: any, func: any, **kwargs) -> any:
    """Applies a function "func" to each "surface" vtkMultiBlockDataSet"""
    if surface.IsA("vtkMultiBlockDataSet"):
        iter = surface.NewIterator()
        iter.UnRegister(None)
        iter.InitTraversal()
        while not iter.IsDoneWithTraversal():
            cur_input = iter.GetCurrentDataObject()
            # numCells = cur_input.GetNumberOfCells()
            numPts = cur_input.GetNumberOfPoints()
            if numPts > 0:
                calc = DataSet(cur_input)
                pts = PointSet(cur_input)
                func(calc, pts, **kwargs)
            iter.GoToNextItem()
    else:
        calc = DataSet(surface)
        pts = PointSet(surface)
        func(calc, pts, **kwargs)


def cp_profile_wall_from_file(
    file_root: str, slice_normal: tuple, slice_origin: tuple, **kwargs
) -> dict:
    """Return chordwise cp profile from wall file"""

    wall = pvs.PVDReader(FileName=file_root)
    clean = clean_vtk(wall)

    pvs.Delete(wall)
    del wall
    profile = cp_profile(clean, slice_normal, slice_origin, **kwargs)
    pvs.Delete(clean)
    del clean

    return profile


def slice_metric(
    surface: any,
    slice_normal: tuple,
    slice_origin: tuple,
    metric_name: str,
    time_average: bool = False,
    **kwargs,
) -> dict:
    """Calculate metric profile from a VTK surface data slice

    Inputs:
    surface: VTKPointData object to take cp cut from
    slice_normal: [nx, ny, nz] normal vector to the cut
    slice_origin: [x, y, z] location of the slice
    metric_name: str name of the metric to be plotted

    Optional:
    time_average: whether to take the time averaged solution for unsteady simulations (default=False)

    Returns:
    slice_dict: dict = {multiblockID: {"chord", "metric"}}
    """

    slice = pvs.Slice(Input=surface, SliceType="Plane")

    slice.SliceType.Normal = slice_normal
    slice.SliceType.Origin = slice_origin

    slice.UpdatePipeline()

    if time_average:
        temporal = pvs.TemporalStatistics(Input=slice)
        temporal.ComputeMaximum = 0
        temporal.ComputeStandardDeviation = 0
        temporal.ComputeMinimum = 0
        temporal.UpdatePipeline()
        pvs.Delete(slice)
        del slice
        slice = temporal

    offset = get_chord(slice)

    chord_calc = pvs.Calculator(Input=slice)
    chord_calc.AttributeType = "Point Data"
    chord_calc.Function = (
        "(coords.iHat - " + str(offset[0]) + ")/" + str(offset[1] - offset[0])
    )
    chord_calc.ResultArrayName = "chord"

    sorted_line = pvs.PlotOnSortedLines(Input=chord_calc)
    sorted_line.UpdatePipeline()

    rawData = pvs.servermanager.Fetch(sorted_line)
    slice_dict = {}
    for block_index in range(rawData.GetNumberOfBlocks()):
        data = DataSet(rawData.GetBlock(block_index))
        slice_dict[block_index] = {
            metric_name: data.PointData[metric_name],
            "chord": data.PointData["chord"],
        }

    # clean up
    pvs.Delete(chord_calc)
    del chord_calc
    pvs.Delete(slice)
    del slice
    pvs.Delete(sorted_line)
    del sorted_line

    return slice_dict


def cp_profile(surface, normal, slice_loc) -> dict:
    """Calculate cp profile from a VTK surface data slice

    Inputs:
    surface: VTKPointData object to take cp cut from
    slice_normal: [nx, ny, nz] normal vector to the cut
    slice_origin: [x, y, z] location of the slice
    metric_name: str name of the metric to be plotted

    Optional:
    time_average: whether to take the time averaged solution for unsteady simulations (default=False)

    Returns:
    slice_dict: dict = {multiblockID: {"chord", "cp"}}
    """
    return slice_metric(surface, normal, slice_loc, "cp")


def cf_profile_wall_from_file(
    file_root: tuple, slice_normal: tuple, slice_origin: tuple, **kwargs
) -> dict:
    """Force coefficient calculation at slice loaction for a file string"""
    wall = pvs.PVDReader(FileName=file_root + "_wall.pvd")
    clean = pvs.CleantoGrid(Input=wall)
    clean.UpdatePipeline()
    inp = pvs.servermanager.Fetch(clean)
    if inp.IsA("vtkMultiBlockDataSet"):
        inp = pvs.MergeBlocks(Input=clean)
    else:
        inp = clean

    pvs.Delete(wall)
    del wall
    out = cf_profile(inp, slice_normal, slice_origin, **kwargs)
    pvs.Delete(clean)
    del clean
    pvs.Delete(inp)
    del inp

    return out


def cf_profile(surface: tuple, slice_normal: tuple, slice_loc: tuple, **kwargs) -> None:
    '''Calculate the force coefficient profile for a slice defined with "slice_normal" and "slice_origin"'''

    cf_calc = pvs.Calculator(Input=surface)

    cf_calc.AttributeType = "Point Data"
    cf_calc.Function = "mag(cf)"
    cf_calc.ResultArrayName = "cfmag"

    return slice_metric(cf_calc, slice_normal, slice_loc, "cfmag")


def screenshot() -> None:
    """TODO finalise..."""
    # position camera
    view = pvs.GetActiveView()
    if not view:
        # When using the ParaView UI, the View will be present, not otherwise.
        view = pvs.CreateRenderView()
    view.CameraViewUp = [0, 0, 1]
    view.CameraFocalPoint = [0, 0, 0]
    view.CameraViewAngle = 45
    view.CameraPosition = [5, 0, 0]

    # draw the object
    pvs.Show()

    # set the background color
    view.Background = [1, 1, 1]  # white

    # set image size
    view.ViewSize = [200, 300]  # [width, height]

    dp = pvs.GetDisplayProperties()

    # set point color
    dp.AmbientColor = [1, 0, 0]  # red

    # set surface color
    dp.DiffuseColor = [0, 1, 0]  # blue

    # set point size
    dp.PointSize = 2

    # set representation
    dp.Representation = "Surface"

    pvs.Render()

    # save screenshot
    pvs.WriteImage("test.png")


def sum_array(input: any, array_name: str) -> list:
    '''Sum of desired array over every cell in "input"'''
    sum = [0.0, 0.0, 0.0]
    p = input.GetCellData().GetArray(array_name)
    numCells = input.GetNumberOfCells()
    for x in range(numCells):
        v = p.GetTuple(x)
        for i in range(0, 3):
            sum[i] += v[i]
    return sum


def get_case_parameters_str(case_name: str, **kwargs) -> Optional[str]:
    "Returns a string of the input dictionary for a zCFD case"
    _data_dir = analysis.data.data_dir
    if "data_dir" in kwargs:
        _data_dir = kwargs["data_dir"]

    try:
        # Get contents of local file
        with open(_data_dir + "/" + case_name + ".py") as f:
            case_file_str = f.read()

            if case_file_str is not None:
                # print status_file_str
                return case_file_str
            else:
                print("WARNING: " + case_name + ".py file not found")
                return None
    except:
        print("WARNING: " + case_name + ".py file not found")
        return None


def get_case_parameters(case_name: str, **kwargs) -> dict:
    """Returns the zCFD input parameter dictionary"""
    case_file_str = get_case_parameters_str(case_name, **kwargs)
    namespace = {}
    exec(case_file_str, namespace)
    return namespace["parameters"]


def get_status_dict(case_name: str, **kwargs) -> Optional[str]:
    """Returns a string of the status of a zCFD run"""
    _data_dir = analysis.data.data_dir
    if "data_dir" in kwargs:
        _data_dir = kwargs["data_dir"]

    try:
        # Get contents of local file
        with open(_data_dir + "/" + case_name + "_status.txt") as f:
            status_file_str = f.read()

            if status_file_str is not None:
                # print status_file_str
                return json.loads(status_file_str)
            else:
                print("WARNING: " + case_name + "_status.txt file not found")
                return None
    except Exception as e:
        print("WARNING: " + case_name + "_status.txt file not found")
        print("Caught exception " + str(e))
        return None


def get_num_procs(case_name: str, **kwargs) -> Optional[int]:
    """Returns the number of processes used in a zCFD run"""
    # remote_host,remote_dir,case_name):
    status = get_status_dict(case_name, **kwargs)
    if status is not None:
        if "num processor" in status:
            return status["num processor"]
        else:
            return None
    else:
        print("status file not found")


def get_case_root(case_name: str, num_procs: Optional[str]) -> str:
    """Returns the output folder path for a specified zCFD run"""
    if num_procs is None:
        num_procs = get_num_procs(case_name)
    return case_name + "_P" + str(num_procs) + "_OUTPUT/" + case_name


def get_case_report(case: str) -> str:
    """Returns the path to the zCFD report file for a specific run"""
    return case + "_report.csv"


def get_case_success(case: str) -> bool:
    """Scans the zCFD log file to identify if a run has successfully completed"""
    # remove .py file exension if its provided
    case_clean = clean_name(case)
    try:
        with open(case_clean + ".log", "r") as f:
            for line in f.readlines():
                if "Solver loop finished" in line:
                    return True
    except FileExistsError as e:
        print(e)
        return False
    return False


class ProgressBar(object):
    """Class to display a progress bar when performing rendering opteraions"""

    def __init__(self) -> None:
        self.pbar = tqdm(total=100)

    def __iadd__(self, v):
        self.pbar.update(v)
        return self

    def complete(self):
        self.pbar.close()

    def update(self, i):
        self.pbar.update(i)



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from .converters import vector_from_wind_dir
from .converters import wind_direction
from .converters import vector_from_angle
from .converters import angle_from_vector
from .converters import rotate_vector
from .converters import feet_to_meters
from .converters import pressure_from_alt
from .converters import rankine_to_kelvin as to_kelvin  # legacy compatibility
from .converters import dot
from .converters import mag
from .converters import R_2vect
from .converters import vector_vector_rotate
from .converters import unit_vector
from .converters import angle_between
from .converters import rotation_matrix
from .converters import rotate
from .converters import project_to_plane
from .converters import rpm2radps

from .fileutils import get_parameters_from_file
from .fileutils import include
from .fileutils import get_zone_info
from .fileutils import zCFD_Result
from .fileutils import get_zcfd_result

from .algorithms import gss



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

zCFD BEM implementation
"""

import numpy as np
from zutil.zWind import transforms, read_default
from typing import Optional


class zTurbineBetBase:
    """subclass for blade element propellor model"""

    def __init__(self, model_dict: dict) -> None:
        self.blade_twist = np.array(model_dict["blade twist"])
        self.blade_chord = np.array(model_dict["blade chord"])
        aerofoils = model_dict["aerofoil positions"]

        self.aerofoil_positions = []
        self.aerofoil_names = []

        for a in aerofoils:
            self.aerofoil_positions.append(a[0])
            self.aerofoil_names.append(a[1])

        self.aerofoils = model_dict["aerofoils"]

        self.tip_loss_correction = model_dict["tip loss correction"]
        self.tip_loss_correction_r = model_dict["tip loss correction radius"]

        self.n_blades = model_dict["number of blades"]

        self.kind = model_dict["kind"]

        self.model_type = "Bet base"

        # if self.tip_loss_correction:
        #     self.tip_loss_correction_r = model_dict["tip loss correction radius"]

    def interp_blade_twist(self, val: float) -> float:
        return np.interp(val, self.blade_twist[:, 0], self.blade_twist[:, 1])

    def interp_blade_chord(self, val: float) -> float:
        return np.interp(val, self.blade_chord[:, 0], self.blade_chord[:, 1])

    def interp_aerofoil_cl(self, val: float, pos: float) -> float:
        bindex = np.digitize(pos, self.aerofoil_positions, right=False)
        aerofoil = self.aerofoil_names[bindex]

        return np.interp(
            val,
            np.array(self.aerofoils[aerofoil]["cl"])[:, 0],
            np.array(self.aerofoils[aerofoil]["cl"])[:, 1],
        )

    def interp_aerofoil_cd(self, val: float, pos: float) -> float:
        bindex = np.digitize(pos, self.aerofoil_positions, right=False)
        aerofoil = self.aerofoil_names[bindex]

        return np.interp(
            val,
            np.array(self.aerofoils[aerofoil]["cd"])[:, 0],
            np.array(self.aerofoils[aerofoil]["cd"])[:, 1],
        )

    def get_tip_loss_factor(self, radial_position: float, turbine: object) -> float:
        if self.tip_loss_correction == "elliptic":
            tlc = np.sqrt(1.0 - (radial_position) ** 2)
        elif self.tip_loss_correction == "acos-fit":
            tlc = (2.0 / np.pi) * np.arccos(
                np.exp(-63.0 * (1.0 - (radial_position) ** 2.0))
            )
        elif self.tip_loss_correction == "acos shift-fit":
            tlc = (2.0 / np.pi) * np.arccos(
                np.exp(-48.0 * (1.0 - (radial_position) ** 2.0) - 0.5)
            )
        elif self.tip_loss_correction == "f-fit":
            tlc = (
                1.0
                - 2.5
                * ((1.0 - (radial_position) ** 2) ** 0.39)
                / (2.0 - (radial_position) ** 2) ** 64
            )
        elif self.tip_loss_correction == "rstar":
            if radial_position > self.tip_loss_correction_r:
                tlc = np.sqrt(
                    1.0
                    - (
                        (radial_position - self.tip_loss_correction_r)
                        / (1 - self.tip_loss_correction_r)
                    )
                    ** 2.0
                )
            else:
                tlc = 1.0
            return tlc
        else:
            return 1.0


class zTurbineBetProp(zTurbineBetBase):
    def __init__(self, control_dict: dict) -> None:
        super().__init__(control_dict)

    def get_sectional_AD_performance(
        self, u_ref: float, density: float, seg: object
    ) -> tuple[float, float]:
        # calculate local velocity vector
        if abs((seg.rp * self.omega) - seg.v_r) > 0.0:
            theta_rel = np.arctan(seg.v_n / ((seg.rp * self.omega) - seg.v_r))
        else:
            theta_rel = np.pi / 2.0

        u_rel = np.sqrt((seg.rp * self.omega - seg.v_r) ** 2 + seg.v_n**2)
        beta_twist = self.interp_blade_twist(seg.rp / self.outer_radius)
        chord = self.interp_blade_chord(seg.rp / self.outer_radius)

        beta = np.deg2rad(beta_twist)
        alpha = beta - theta_rel

        cl = self.interp_aerofoil_cl(alpha, seg.rp / self.outer_radius)
        cd = self.interp_aerofoil_cd(alpha, seg.rp / self.outer_radius)

        if self.tip_loss_correction:
            tlc = self.get_tip_loss_factor(seg)
            cl = cl * tlc
            cd = cd * tlc

        f_L = cl * 0.5 * density * u_rel**2 * chord
        f_D = cd * 0.5 * density * u_rel**2 * chord

        F_L = (self.n_blades / (2.0 * np.pi * seg.rp)) * f_L
        F_D = (self.n_blades / (2.0 * np.pi * seg.rp)) * f_D

        dt = -(F_L * np.cos(theta_rel) - F_D * np.sin(theta_rel)) * seg.da
        dq = -(F_L * np.sin(theta_rel) - F_D * np.cos(theta_rel)) * seg.da

        return dt, dq


class zTurbineBetFull(zTurbineBetBase):
    def __init__(self, model_dict: dict) -> None:
        super().__init__(model_dict)

        self.n_sections = model_dict["number of sections"]

        self.sections = np.linspace(0, 1, self.n_sections)

    def get_performance(self, u_ref: float, turbine: object, density: float) -> None:
        """Main body of BEMT code- work along each blade section, calculate the induced velocity vector at each section, then calculate sectional forces from polars"""

        # iteratate along span of blade
        for s in self.sections:
            # at each section caculate the induced velocity vector
            pass

    def discrete_performance(
        self, u_ref: float, density: float, seg: object, turbine: object
    ) -> tuple[float, float]:
        """takes an input of the desired rotor pitch angle, and returns the predicted performance of the turbine"""
        omega_rel = turbine.controller.omega - seg.omega_air
        u_rel = np.sqrt((seg.rp * omega_rel) ** 2 + seg.u_ref_local**2)

        if (seg.rp * omega_rel) > 0.0:
            theta_rel = np.arctan(seg.u_ref_local / (seg.rp * omega_rel))
        else:
            theta_rel = np.pi / 2.0

        radial_position = seg.rp / turbine.outer_radius

        beta_twist = self.interp_blade_twist(radial_position)
        chord = self.interp_blade_chord(radial_position) * turbine.outer_radius

        if self.kind == "propellor":
            theta_rel *= -1  # invert for sign convention

        beta = np.deg2rad(turbine.controller.pitch + beta_twist)
        alpha = theta_rel - beta

        if self.kind == "propellor":
            alpha *= -1  # invert for debugging

        cl = self.interp_aerofoil_cl(np.rad2deg(alpha), radial_position)

        tlc = self.get_tip_loss_factor(radial_position, turbine)
        cl *= tlc

        cd = self.interp_aerofoil_cd(np.rad2deg(alpha), radial_position)
        cd *= tlc

        f_L = cl * 0.5 * density * u_rel**2 * chord
        f_D = cd * 0.5 * density * u_rel**2 * chord

        F_L = (self.n_blades / (2.0 * np.pi * seg.rp)) * f_L
        F_D = (self.n_blades / (2.0 * np.pi * seg.rp)) * f_D

        dt = (F_L * np.cos(theta_rel) - F_D * np.sin(theta_rel)) * seg.da
        dq = (F_L * np.sin(theta_rel) + F_D * np.cos(theta_rel)) * seg.da

        if self.kind == "propellor":
            dq *= -1
            dt *= -1
        if turbine.rotation_direction == "anticlockwise":
            dq *= -1

        return dt, dq

    def init_bet_kernel(self, u_ref: float, density: float, seglist: list) -> None:
        pass

    def bet_kernel(self, pitch: float) -> None:
        pass


class Blade_Element:
    def __init__(
        self,
        p1: np.array,
        p2: np.array,
        chord: float,
        twist: float,
        aerofoil: str,
        verbose: bool = False,
    ) -> None:
        self.verbose = verbose
        # points defining each end of element
        self.p1 = p1
        self.p2 = p2

        # vector from element centre to centre of rotation
        self.r0 = (p2 + p1) / 2

        # aerodynamic properties
        self.chord = chord
        self.aerofoil = aerofoil

        # angular properties of each element
        self.twist = twist

        p12 = self.p2 - self.p1

        # figure these two out geometrically
        self.sweep = np.arctan(p12[1] / p12[0])
        self.prebend = -np.arctan(p12[1] / (np.sqrt(p12[0] ** 2 + p12[1] ** 2)))

        self.calculate_induced_angular_velocity()

        self.T_bc = transforms.T_bc(
            np.deg2rad(self.twist), np.deg2rad(self.sweep), np.deg2rad(self.prebend)
        )

        # figure out triad for specific element

    def calculate_induced_angular_velocity(self) -> None:
        omega = np.array([0.0, 0.79, 0.0])
        self.phi_induced = -np.cross(omega, self.r0)

    def global2element(self, V: np.array) -> None:
        return


class BEMT_Blade:
    def __init__(self, model_dict: dict, azimuth: float, verbose: bool = False) -> None:
        self.verbose = verbose

        # angular properties of each blade
        self.cone = read_default(model_dict, "cone", 0.0)
        self.pitch = 0.0
        self.num_elements = model_dict["number of elements"]
        self.azimuth = azimuth

        self.blade_axis = np.array(model_dict["blade axis"])
        self.blade_axis[:, 3] += 3.0
        # swap x and z to coincide with atom axis (FOR NOW)

        self.blade_axis[:, [1, 3, 2]] = self.blade_axis[:, [3, 2, 1]]

        self.chord = np.array(model_dict["blade chord"])
        self.twist = np.array(model_dict["blade twist"])

        aerofoils = model_dict["aerofoil positions"]

        self.aerofoil_positions = []
        self.aerofoil_names = []

        for a in aerofoils:
            self.aerofoil_positions.append(a[0])
            self.aerofoil_names.append(a[1])

        self.T_hb = transforms.T_hb(
            np.deg2rad(self.azimuth), np.deg2rad(self.cone), np.deg2rad(self.pitch)
        )

        # create elemments
        if self.verbose:
            print("discretising blade with {} elements".format(self.num_elements))
        self.elements = []
        self.nodes = np.linspace(0, 1.0, self.num_elements + 1)
        for i in range(self.num_elements):
            # points at either end of element
            p1 = self.get_blade_axis_point(self.nodes[i])
            p2 = self.get_blade_axis_point(self.nodes[i + 1])

            # chord, twist, aerofoil at midpoint
            chord = self.interp_blade_chord((self.nodes[i] + self.nodes[i + 1]) / 2)
            twist = self.interp_blade_twist((self.nodes[i] + self.nodes[i + 1]) / 2)
            aerofoil = self.get_aerofoil((self.nodes[i] + self.nodes[i + 1]) / 2)
            self.elements.append(Blade_Element(p1, p2, chord, twist, aerofoil))

    def get_blade_axis_point(self, radial_position: float) -> list:
        # returns the interpolated point on the blade axis at point `radial_position`
        point = np.array([0.0, 0.0, 0.0])
        for i, _ in enumerate(point):
            point[i] = np.interp(
                radial_position, self.blade_axis[:, 0], self.blade_axis[:, i + 1]
            )
        return point

    def interp_blade_chord(self, val: float) -> float:
        return np.interp(val, self.chord[:, 0], self.chord[:, 1])

    def interp_blade_twist(self, val: float) -> float:
        return np.interp(val, self.twist[:, 0], self.twist[:, 1])

    def get_aerofoil(self, pos: float) -> str:
        bindex = np.digitize(pos, self.aerofoil_positions, right=False)
        aerofoil = self.aerofoil_names[bindex]
        return aerofoil

    def create_elements(self) -> None:
        pass

    def get_induced_velocities(self) -> np.array:
        induced_velocities = np.zeros((self.num_elements, 3))
        for i, e in enumerate(self.elements):
            induced_velocities[i, :] = e.phi_induced

        return induced_velocities

    def blade2hub(self, V: np.array) -> np.array:
        # transform points in the blade coordinate system to the hub coordinate system
        return transforms.R_yy(np.deg2rad(self.azimuth)) @ V

    def blade2global(self, V: np.array) -> None:
        return


class BEMT_Turbine:
    def __init__(self, model_dict: dict, verbose: bool = False) -> None:
        self.verbose = verbose

        self.num_blades = model_dict["number of blades"]
        self.yaw = read_default(model_dict, "yaw", 0.0)
        self.tilt = read_default(model_dict, "tilt", 0.0)
        self.azimuth_offset = read_default(model_dict, "azimuth", 0.0)

        self.T_gh = transforms.T_gh(
            np.deg2rad(self.yaw), np.deg2rad(self.tilt), np.deg2rad(self.azimuth_offset)
        )

        # create blades
        self.blades = []
        if self.verbose:
            print("Creating {} blades".format(self.num_blades))
        azimuth_positions = (
            np.linspace(0, 360, self.num_blades + 1) + self.azimuth_offset
        )
        for i in range(self.num_blades):
            self.blades.append(BEMT_Blade(model_dict, azimuth_positions[i]))

        self.model_type = "BEMT Turbine"

    def discrete_performance(
        self, u_ref: float, density: float, seg: object, turbine: object
    ) -> tuple:
        return 0.0, 0.0

    def plot_turbine(self) -> None:
        # util for visualising the turbine

        nodes = []
        elements = []
        rp = []
        mechanical_velocities = []
        local_velocities = []
        i = 0
        for blade in self.blades:
            for e in blade.elements:
                elements.append([i, i + 1])

                # coordinates in blade axis
                triad = e.T_bc @ blade.T_hb @ self.T_gh
                p = np.linalg.inv(triad) @ e.p1
                nodes.append(p)

                r = np.linalg.inv(triad) @ e.r0
                rp.append(r)

                mechanical_velocity = np.linalg.inv(triad) @ e.phi_induced
                mechanical_velocities.append(mechanical_velocity)

                local_velocity = mechanical_velocity + np.array([0.0, 10.59, 0.0])
                local_velocities.append(local_velocity)

                i += 1
            nodes.append(np.linalg.inv(triad) @ e.p2)
            i += 1

        cell_types = np.array([3 for i in range(len(elements))])

        point_data = {}

        point_data["scalars"] = {}
        point_data["scalars"]["id"] = [i for i in range(len(nodes))]

        point_data["vectors"] = {}
        point_data["vectors"]["normal"] = [
            [float(i), float(i), float(i)] for i in range(len(nodes))
        ]

        element_data = {}
        element_data["scalars"] = {}
        element_data["scalars"]["id"] = [i for i in range(len(elements))]

        element_data["vectors"] = {}
        element_data["vectors"]["normal"] = [[i, i, i] for i in range(len(elements))]
        element_data["vectors"]["rp"] = rp
        element_data["vectors"]["mechanical_velocity"] = mechanical_velocities
        element_data["vectors"]["local_velocity"] = local_velocities

        write_vtk_fe_mesh(
            nodes,
            elements,
            cell_types,
            "test.vtk",
            point_data=point_data,
            element_data=element_data,
        )


def write_vtk_fe_mesh(
    nodes: list,
    elements: list,
    cell_types: list,
    file_path: str,
    point_data: Optional[dict] = None,
    element_data: Optional[dict] = None,
) -> None:
    with open(file_path, "w") as vtk_file:
        vtk_file.write("# vtk DataFile Version 4.2\n")
        vtk_file.write("FE Mesh Data\n")
        vtk_file.write("ASCII\n")
        vtk_file.write("DATASET UNSTRUCTURED_GRID\n")

        # Write the node coordinates
        vtk_file.write(f"POINTS {len(nodes)} double\n")
        for node in nodes:
            vtk_file.write(f"{node[0]} {node[1]} {node[2]}\n")

        # Write the element connectivity
        num_elements = len(elements)
        total_num_entries = sum(
            len(cell) + 1 for cell in elements
        )  # Sum of nodes + 1 for the cell type
        vtk_file.write(f"\nCELLS {num_elements} {total_num_entries}\n")
        for element in elements:
            num_nodes = len(element)
            vtk_file.write(f"{num_nodes} {' '.join(str(node) for node in element)}\n")

        # Write the cell types
        vtk_file.write(f"\nCELL_TYPES {num_elements}\n")
        for cell_type in cell_types:
            vtk_file.write(f"{cell_type}\n")

        if point_data:
            vtk_file.write(f"\nPOINT_DATA {len(nodes)}\n")
            if "scalars" in point_data.keys():
                for pd in point_data["scalars"].keys():
                    vtk_file.write(f"\nSCALARS {pd} double 1\n")
                    vtk_file.write("LOOKUP_TABLE default\n")
                    for point in point_data["scalars"][pd]:
                        vtk_file.write(f"{point}\n")

            if "vectors" in point_data.keys():
                for vd in point_data["vectors"].keys():
                    vtk_file.write(f"\nVECTORS {vd} double\n")
                    for vector in point_data["vectors"][vd]:
                        for v in vector:
                            vtk_file.write(f"{v} ")
                        vtk_file.write("\n")

        if element_data:
            vtk_file.write(f"\nCELL_DATA {len(elements)}\n")
            if "scalars" in element_data.keys():
                for pd in element_data["scalars"].keys():
                    vtk_file.write(f"\nSCALARS {pd} double 1\n")
                    vtk_file.write("LOOKUP_TABLE default\n")
                    for point in element_data["scalars"][pd]:
                        vtk_file.write(f"{point}\n")

            if "vectors" in element_data.keys():
                for vd in element_data["vectors"].keys():
                    vtk_file.write(f"\nVECTORS {vd} double\n")
                    for vector in element_data["vectors"][vd]:
                        for v in vector:
                            vtk_file.write(f"{v} ")
                        vtk_file.write("\n")

    print(f"VTK FE mesh has been saved to {file_path}")



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Converter utilities
"""

import numpy as np


def vector_from_wind_dir(wind_dir_degree: float, wind_speed: float = 1.0) -> list:
    """
    Return vector given a wind direction and wind speed
    Wind dir = 0.0 -> [0.0,-1.0,0.0]
    Wind dir = 90.0 -> [-1.0,0.0,0.0]

    Wind dir - Meteorological wind direction
    (direction from which wind is blowing)

    u -> Zone Velocity (Towards East)
    v -> Meridional Velocity (Towards North)

    """

    return [
        -wind_speed * np.sin(np.deg2rad(wind_dir_degree)),
        -wind_speed * np.cos(np.deg2rad(wind_dir_degree)),
        0.0,
    ]


def wind_direction(uvel: float, vvel: float) -> float:
    """
    Calculate meteorological wind direction from velocity vector
    """
    return np.rad2deg(np.arctan2(-uvel, -vvel))


def vector_from_angle(alpha: float, beta: float, mag: float = 1.0) -> list:
    """
    Return vector given alpha and beta in degrees based on ESDU definition
    """
    alpha = np.deg2rad(alpha)
    beta = np.deg2rad(beta)
    vec = [0.0, 0.0, 0.0]
    vec[0] = mag * np.cos(alpha) * np.cos(beta)
    vec[1] = mag * np.sin(beta)
    vec[2] = mag * np.sin(alpha) * np.cos(beta)
    return vec


def angle_from_vector(vec: list) -> tuple[float, float]:
    """
    Return vector given alpha and beta in degrees based on ESDU definition
    """
    mag = np.sqrt(vec[0] * vec[0] + vec[1] * vec[1] + vec[2] * vec[2])

    beta = np.arcsin(vec[1] / mag)
    alpha = np.arccos(vec[0] / (mag * np.cos(beta)))
    alpha = np.rad2deg(alpha)
    beta = np.rad2deg(beta)
    return (alpha, beta)


def rotate_vector(vec: list, alpha_degree: float, beta_degree: float) -> list:
    """
    Rotate vector by alpha and beta based on ESDU definition
    """
    alpha = np.deg2rad(alpha_degree)
    beta = np.deg2rad(beta_degree)
    rot = [0.0, 0.0, 0.0]
    rot[0] = (
        np.cos(alpha) * np.cos(beta) * vec[0]
        + np.sin(beta) * vec[1]
        + np.sin(alpha) * np.cos(beta) * vec[2]
    )
    rot[1] = (
        -np.cos(alpha) * np.sin(beta) * vec[0]
        + np.cos(beta) * vec[1]
        - np.sin(alpha) * np.sin(beta) * vec[2]
    )
    rot[2] = -np.sin(alpha) * vec[0] + np.cos(alpha) * vec[2]
    return rot


def feet_to_meters(val: float) -> float:
    """Convert feet to meters"""
    return val * 0.3048


def pressure_from_alt(alt: float) -> float:
    """
    Calculate pressure in Pa from altitude in m using standard atmospheric tables
    """
    return 101325.0 * (1.0 - 2.25577e-5 * alt) ** 5.25588


def rankine_to_kelvin(rankine: float) -> float:
    """Convert ranking to kelvin"""
    return rankine * 0.555555555


def dot(vec1: list, vec2: list) -> float:
    """Returns the dot product of two vectors - to be depreciated in favour of just using np.dot"""
    return np.dot(vec1, vec2)


def mag(vec: list) -> float:
    """Return the magnitude of a vector- to be depreciated in favour of just using np.linalg.norm"""
    return np.linalg.norm(vec)


def R_2vect(vector_orig: np.ndarray, vector_fin: np.ndarray) -> None:
    """
    Calculates the rotation matrix required to rotate from one vector to another.

    Args:
        R: The 3x3 rotation matrix to update (3x3 NumPy array).
        vector_orig: The unrotated vector in the reference frame (3x1 NumPy array).
        vector_fin: The rotated vector in the reference frame (3x1 NumPy array).
    """
    R = np.zeros((3, 3))
    # Normalize vectors and handle degenerate cases
    unit_orig = vector_orig / np.linalg.norm(vector_orig)
    unit_fin = vector_fin / np.linalg.norm(vector_fin)
    if np.allclose(unit_orig, unit_fin):
        # Vectors are equal or opposite, no rotation needed
        R[:] = np.eye(3)
        return R

    # Calculate and normalize rotation axis
    axis = np.cross(unit_orig, unit_fin)
    axis_len = np.linalg.norm(axis)
    axis = axis / axis_len if axis_len else np.zeros_like(axis)

    # Calculate rotation angle and trigonometric functions
    angle = np.arccos(np.dot(unit_orig, unit_fin))
    ca = np.cos(angle)
    sa = np.sin(angle)

    # Alias the axis coordinates.
    x = axis[0]
    y = axis[1]
    z = axis[2]

    R[0, 0] = 1.0 + (1.0 - ca) * (x**2 - 1.0)
    R[0, 1] = -z * sa + (1.0 - ca) * x * y
    R[0, 2] = y * sa + (1.0 - ca) * x * z
    R[1, 0] = z * sa + (1.0 - ca) * x * y
    R[1, 1] = 1.0 + (1.0 - ca) * (y**2 - 1.0)
    R[1, 2] = -x * sa + (1.0 - ca) * y * z
    R[2, 0] = -y * sa + (1.0 - ca) * x * z
    R[2, 1] = x * sa + (1.0 - ca) * y * z
    R[2, 2] = 1.0 + (1.0 - ca) * (z**2 - 1.0)
    return R


def vector_vector_rotate(
    vec: np.ndarray, axis: np.ndarray, origin: np.ndarray, theta: float
) -> np.ndarray:
    """
    Rotates a vector around an arbitrary axis passing through an origin.

    Args:
        vec: The vector to be rotated (3x1 NumPy array).
        axis: The rotation axis (3x1 NumPy array).
        origin: The point through which the axis passes (3x1 NumPy array).
        theta: The rotation angle in radians.

    Returns:
        The rotated vector (3x1 NumPy array).
    """

    # Ensure inputs are NumPy arrays
    vec, axis, origin = np.asarray(vec), np.asarray(axis), np.asarray(origin)

    # Normalize the axis vector
    axis = axis / np.linalg.norm(axis)

    # Calculate intermediate terms for efficiency
    c = np.cos(theta)
    s = np.sin(theta)
    t = 1 - c
    v = vec - origin
    v_proj = np.dot(v, axis) * axis  # Projection of v onto axis

    # Perform rotation using vectorized operations
    rotated_vec = (
        v_proj * t
        + v * c  # Unrotated component along axis
        + np.cross(axis, v)  # Rotated component perpendicular to axis
        * s  # Rotation around axis
        + origin  # Translate back to original origin
    )

    return rotated_vec


def unit_vector(vector: list) -> list:
    """Return the unit vector in the direction of the original vector"""
    return vector / np.linalg.norm(vector)


def angle_between(v1: list, v2: list) -> float:
    """Return the interior angle between two vectors"""
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def rotation_matrix(axis: list, theta: float) -> np.array:
    """Return the rotation matrix required to rotate a point about a specied axis, by a specified angle"""
    axis = np.asarray(axis)
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(theta / 2.0)
    b, c, d = -axis * np.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array(
        [
            [aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
            [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
            [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc],
        ]
    )


def rotate(coord: list, axis: list, ang: float) -> list:
    """Rotate a coordinate about a specified axis, by a specified angle"""
    c2 = np.dot(rotation_matrix(axis, ang), (coord[0], coord[1], coord[2]))
    return [c2[0], c2[1], c2[2]]


def project_to_plane(
    pt: np.array, plane_point: np.array, plane_normal: np.array
) -> np.array:
    """
    Projects a point onto a plane defined by a point and a normal vector.

    Args:
        pt (np.ndarray): The 3D point to be projected (shape: (3,)).
        plane_point (np.ndarray): A point on the plane (shape: (3,)).
        plane_normal (np.ndarray): The normal vector of the plane (shape: (3,)).

    Returns:
        np.ndarray: The projected point (shape: (3,)).
    """

    # Calculate the vector from the plane point to the input point
    v = pt - plane_point

    # Project the vector onto the plane normal
    projection = np.dot(v, plane_normal) * plane_normal

    # Subtract the projection from the vector to get the projected point
    projected_point = pt - projection

    return projected_point


def rpm2radps(rpm: float) -> float:
    """Converts rotations per minute to radians per second"""
    return 2 * np.pi * rpm / 60


def radps2rps(radps: float) -> float:
    """Converts radians per second to rotations per second.

    Needed for evaluation of propellor thrust coefficient"""
    return radps / (2 * np.pi)



"""
Copyright (c) 2012-2025, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Helper functions for plotting FWH results
"""

import matplotlib.pyplot as plt
from typing import Optional, Union
from zutil.plot import (
    get_figure,
    x_label,
    y_label,
    set_title,
    set_legend,
    plt_logo_stamp,
)
from collections.abc import Iterable
import pandas as pd

from zutil.analysis.acoustic import (
    calculate_PSD,
    calculate_thirdoctave_bands,
    convert_to_dB,
)


def plot_thirdoctave(
    p: list,
    time: Union[Iterable, float],
    sampling_frequency: float = 1.0,
    title: str = "Third Octave Band Data",
    label: Optional[str] = None,
    ax: Optional[plt.Axes] = None,
) -> plt.Axes:
    """
    Plots third-octave band data for given observer data.

    Args:
        p (Arraylike): Pressure data array
        time [list, float]: time array, or timestep
        sampling_frequency (float = 1.0): Sampling frequency in Hz.
    Optional:
        title (str): Title of the plot.
        label (str): Label for the plot.
        ax (plt.Axes): Matplotlib Axes object to plot on. If None, a new figure is created.
    """
    # Initialize figure and set labels

    if ax is None:
        # create new figure if no axes are provided
        print("Creating new figure for third-octave plot")
        fig = get_figure()
        ax = fig.gca()
        ax.grid(True)

    else:
        fig = ax.figure

    x_label(ax, "Frequency [Hz]")
    y_label(ax, "dB [1/3 Octave]")

    # calculate Power Spectral Density (PSD) using Welch's method
    f1, Pxx_den1 = calculate_PSD(p, time, sampling_frequency=sampling_frequency)
    third_octave_freq, third_octave_data = calculate_thirdoctave_bands(f1, Pxx_den1)
    third_octave_data_db = convert_to_dB(third_octave_data)

    # Plot third-octave data for the current lifter
    ax.semilogx(third_octave_freq, third_octave_data_db, label=label)

    # Finalize plot
    pos = ax.get_position()
    ax.set_position([pos.x0, pos.y0, pos.width * 0.8, pos.height])

    if label:
        # Add legend if label is provided
        set_legend(ax, "upper right")
    if title:
        # Set plot title if provided
        set_title(ax, title)

    return ax


def plot_observer(
    p: list,
    T: list,
    title: str = "Observer Data",
    label: Optional[str] = None,
    ax: Optional[plt.Axes] = None,
) -> plt.Axes:
    """
    Plots observer data for given observer data.

    Args:
        p (Arraylike): Pressure data array
        T (Arraylike): Time data array
    Optional:
        title (str): Title of the plot.
        label (str): Label for the plot.
        ax (plt.Axes): Matplotlib Axes object to plot on. If None, a new figure is created.
    """
    # Initialize figure and set labels
    if ax is None:
        # create new figure if no axes are provided
        fig = get_figure()
        ax = fig.gca()
        ax.grid(True)

    else:
        fig = ax.figure

    # Set plot title
    if title:
        set_title(ax, title)

    x_label(ax, "Time [s]")
    y_label(ax, "Pressure [Pa]")

    # Plot observer data for the current lifter
    ax.plot(T, p, label=label)

    # Finalize plot
    pos = ax.get_position()
    ax.set_position([pos.x0, pos.y0, pos.width * 0.8, pos.height])

    if label:
        set_legend(ax, "upper right")

    return ax


def plot_PSD(
    p: list,
    T: list,
    sampling_frequency: float = 1.0,
    title: str = "PSD",
    label: Optional[str] = None,
    ax: Optional[plt.Axes] = None,
) -> plt.Axes:
    """
    Plots Power Spectral Density (PSD) for given observer data.

    Args:
        p (Arraylike): Pressure data array
        T (Arraylike): Time data array
        sampling_frequency (float = 1.0): Sampling frequency in Hz.
    Optional:
        title (str): Title of the plot.
        label (str): Label for the plot.
        ax (plt.Axes): Matplotlib Axes object to plot on. If None, a new figure is created.
    """
    # Initialize figure and set labels
    if ax is None:
        # create new figure if no axes are provided
        fig = get_figure()
        ax = fig.gca()
        ax.grid(True)
        print("Creating new figure for PSD plot")
    else:
        fig = ax.figure

    # Set plot title
    if title:
        set_title(ax, title)

    x_label(ax, "Frequency [Hz]")
    y_label(ax, "PSD [dB]")

    plt.ylim(0, 100)  # Set y-axis limits for PSD plot

    # Compute and plot Power Spectral Density (PSD) for p
    f, Pxx_den = calculate_PSD(p, T, sampling_frequency)
    Pxx_den_db = convert_to_dB(Pxx_den)

    ax.semilogx(f, Pxx_den_db, label=label)  # Plot on a logarithmic scale

    if label:
        # Add legend if label is provided
        set_legend(ax, "upper right")

    set_title(ax, title)

    return ax


def plot_all_PSD(
    filepath: str, sampling_frequency: float = 1.0, title: str = "PSD", stamp=False
) -> None:
    """
    Plot all PSDs from a given observer file

    Args:
        filepath (str): Path to the file containing observer data.
        sampling_frequency (float): Sampling frequency in Hz.
    """
    # Read the file and extract data
    data = pd.read_csv(filepath)

    # create new figure
    fig = get_figure()
    ax = fig.gca()
    ax.grid(True)

    time = data["Time"].to_list()
    # Loop through each observer and plot its PSD
    for column in data.columns:
        if column != "Time":
            p = data[column].to_list()
            # Plot PSD for the current observer
            plot_PSD(
                p,
                time,
                sampling_frequency=sampling_frequency,
                label=column,
                ax=ax,
                title=title,
            )

    if stamp:
        plt_logo_stamp(ax, location=(0.1, 0.9))
    fig.tight_layout()
    plt.legend()
    plt.show()


def plot_all_thirdoctave(
    filepath: str,
    sampling_frequency: float = 1.0,
    title: str = "Third Octave Band Data",
    stamp=False,
) -> None:
    """
    Plot all third-octave bands from a given observer file

    Args:
        filepath (str): Path to the file containing observer data.
        sampling_frequency (float): Sampling frequency in Hz.
    """
    # Read the file and extract data
    data = pd.read_csv(filepath)

    # create new figure
    fig = get_figure()
    ax = fig.gca()
    ax.grid(True)

    time = data["Time"].to_list()
    # Loop through each observer and plot its third-octave band data
    for column in data.columns:
        if column != "Time":
            p = data[column].to_list()
            # Plot third-octave band data for the current observer
            plot_thirdoctave(
                p,
                time,
                sampling_frequency=sampling_frequency,
                label=column,
                ax=ax,
                title=title,
            )
    if stamp:
        plt_logo_stamp(ax, location=(0.1, 0.9))
    fig.tight_layout()

    plt.legend()
    plt.show()



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Module to contain functionality of actuator disc methods
"""

from zutil.zWind.zTurbine import zTurbine
import numpy as np
from zutil import R_2vect
from typing import Optional
import matplotlib.pyplot as plt


class zActuatorDiscSegment:
    """
    Represents an individual actuator disc segment with associated attributes and methods.
    """

    def __init__(
        self,
        r: float,
        dr: float,
        theta: float,
        dtheta: float,
        disc_normal: np.ndarray,
        disc_centre: np.ndarray,
        rotation_direction: str = "clockwise",
    ) -> None:
        self.r = r
        self.dr = dr
        self.theta = theta
        self.dtheta = dtheta
        self.disc_normal = disc_normal
        self.disc_centre = disc_centre
        self.rotation_direction = rotation_direction

        self.rp = r + 0.5 * dr
        self.da = self.dtheta * self.rp * dr
        disc_theta = self.theta + 0.5 * self.dtheta

        # Calculate local_vector directly in global coordinates
        self.local_vector = (
            self.rp
            * np.array([np.cos(disc_theta), np.sin(disc_theta), 0.0])
            @ np.linalg.inv(self._rotation_matrix())
        )

        self.segment_centre = self.disc_centre + self.local_vector

    def _rotation_matrix(self) -> np.ndarray:
        """Constructs the rotation matrix from the disc normal."""
        return R_2vect(np.array([0, 0, 1]), self.disc_normal)

    def calculate_velocity_vector(self, ulocal: np.ndarray) -> None:
        """Calculates local velocity components for the segment."""
        rvec = self.local_vector / np.linalg.norm(self.local_vector)  # get unit vector

        if self.rotation_direction == "clockwise":
            local_omega_vec = np.cross(rvec, self.disc_normal)
        else:
            local_omega_vec = np.cross(self.disc_normal, rvec)

        self.v_n = -np.dot(ulocal, self.disc_normal)
        self.v_r = np.dot(ulocal, local_omega_vec)
        self.omega_air = np.dot(local_omega_vec, ulocal) / self.rp
        self.u_ref_local = np.dot(ulocal, self.disc_normal)


class zActuactorDisc:
    """Main class for handling actuator disc operations"""

    def __init__(self, control_dict: dict, verbose: bool = False) -> None:
        # create turbine model

        self.n_segments = control_dict["discretisation"]["number of elements"]
        self.verbose = verbose

        self.turbine = zTurbine(control_dict)

        self.create_annulus()

    def __enter__(self) -> object:
        return self

    def __exit__(self, type, value, traceback) -> bool:  # noqa: ANN001
        return False

    def create_annulus(self) -> list:
        """Discretise the area of the disc, and store in list of annulus segments"""
        self.annulus = []
        self.dtheta = np.deg2rad(360.0 / self.n_segments)
        for i in range(self.n_segments):
            r = self.turbine.inner_radius
            while r < self.turbine.outer_radius:
                dr = (
                    self.dtheta
                    * max(r, 0.01 * self.turbine.outer_radius)
                    / (1.0 - 0.5 * self.dtheta)
                )
                max_r = r + dr
                if max_r > self.turbine.outer_radius:
                    dr = self.turbine.outer_radius - r

                self.annulus.append(
                    zActuatorDiscSegment(
                        r,
                        dr,
                        i * self.dtheta,
                        self.dtheta,
                        self.turbine.normal,
                        self.turbine.centre,
                    )
                )

                r += dr

        return self.annulus

    def get_annulus(self) -> list:
        """Returns the annulus data in a C++-compatible format."""
        return [
            (
                seg.r,
                seg.dr,
                seg.theta,
                seg.dtheta,
                seg.segment_centre[0],
                seg.segment_centre[1],
                seg.segment_centre[2],
            )
            for seg in self.annulus
        ]

    def create_turbine_segments(
        self,
        u_ref: float,
        density: float,
        annulusVel: Optional[list] = None,
        turbine_name_dict: Optional[dict] = {},
        turbine_name: str = "",
    ) -> list:
        """Main body of actuator disc code- take in velocities at the turbine disc, return the torque and moment at each section"""
        if annulusVel:
            # if annulus_vec is supplied, u_ref = mean of annulus vec cells
            temp = np.reshape(annulusVel, (-1, 3)).T
            self.u_ref = np.sqrt(
                np.mean(temp[0]) ** 2 + np.mean(temp[1]) ** 2 + np.mean(temp[2]) ** 2
            )
        else:
            # take u_ref from code
            self.u_ref = u_ref

        self.density = density

        # reshape list of annular velocities into nx3 vector
        if annulusVel:
            self.annulus_velocities = np.reshape(annulusVel, (-1, 3))
        else:
            self.annulus_velocities = []

        converged = False

        # controller loop- get initial turbine conditions
        pitch, omega = self.turbine.controller.get_conditions()
        while not converged:
            # self.turbine.set_omega(omega)
            # self.turbine.set_pitch(pitch)
            # get performance for current operating conditions
            self.calculate_turbine_performance(
                pitch, turbine_name_dict=turbine_name_dict, turbine_name=turbine_name
            )

            # check for controller convergence
            pitch, omega, converged = self.turbine.controller.update_conditions(
                self.u_ref, self.torque
            )

        return self.annulus_out

    def calculate_turbine_performance(
        self,
        pitch: float,
        turbine_name_dict: Optional[dict] = None,
        turbine_name: str = "",
    ) -> None:
        # get predicted power and torque of turbine

        total_area = 0
        total_thrust = 0
        total_torque = 0

        self.annulus_out = []

        for i, seg in enumerate(self.annulus):
            if len(self.annulus_velocities) > 0:
                seg.calculate_velocity_vector(self.annulus_velocities[i])
            dt, dq = self.turbine.get_discrete_performance(self.u_ref, seg)
            self.annulus_out.append((dt, dq, seg.r, seg.dr, seg.theta, seg.dtheta))

            total_area += seg.da
            total_thrust += dt
            total_torque += np.abs(dq * seg.rp)

        total_power = total_torque * self.turbine.omega

        # Report back to turbine power
        turbine_name_dict[turbine_name + "_power"] = total_power
        turbine_name_dict[turbine_name + "_uref"] = self.u_ref
        turbine_name_dict[turbine_name + "_omega"] = self.turbine.omega
        turbine_name_dict[turbine_name + "_thrust"] = total_thrust

        self.torque = total_torque

    # Visualisation aids
    def plot_annulus(self) -> plt.Axes:
        """ "Plot a wireframe of the returned annulus"""
        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})

        # unpack annulus out vector
        nt = self.n_segments
        num_pts = len(self.annulus_out)
        nr = int(num_pts / nt)

        R = np.zeros((nr, nt))
        T = np.zeros((nr, nt))
        Tr = np.zeros((nr, nt))

        ii = 0

        for t in range(nt):
            for r in range(nr):
                R[r, t] = self.annulus_out[ii][2]
                T[r, t] = self.annulus_out[ii][4]
                Tr[r, t] = self.annulus_out[ii][1]

                ii += 1

        X, Y = R * np.cos(T), R * np.sin(T)

        # Stitch across last row
        X = np.column_stack((X, X[:, 0]))
        Y = np.column_stack((Y, Y[:, 0]))
        Tr = np.column_stack((Tr, Tr[:, 0]))

        ax.plot_wireframe(X, Y, Tr)

        return ax


def convolution(
    disc: list,
    disc_centre: tuple,
    disc_radius: float,
    disc_normal: type,
    disc_up: tuple,
    cell_centre_list: list,
    cell_volume_list: list,
) -> tuple:
    """
    Performs a convolution of discrete disc annulus data to then background CFD mesh.

    Args:
        disc: (list) Representation of the disc.
        disc_centre: (numpy.ndarray) Centre of the disc.
        disc_radius: (float) Radius of the disc.
        disc_normal: (numpy.ndarray) Normal vector of the disc.
        disc_up: (numpy.ndarray) Up vector of the disc.
        cell_centre_list: (list) List of cell centres.
        cell_volume_list: (list) List of cell volumes.

    Returns:
        list: List of scaled cell forces, where each element is a tuple of (x, y, z) forces.
    """

    from mpi4py import MPI
    import libconvolution as cv

    cell_centre_list_np = np.asarray(cell_centre_list)
    cell_volume_list_np = np.asarray(cell_volume_list)

    weighted_sum = np.zeros(len(disc))

    weighted_sum = cv.convolution_2dkernel_weights(
        disc,
        disc_centre,
        disc_radius,
        disc_normal,
        disc_up,
        cell_centre_list_np,
        cell_volume_list_np,
    )
    # Need to reduce weighted sum over all processes
    totals = np.zeros_like(weighted_sum)

    MPI.COMM_WORLD.Allreduce(weighted_sum, totals, op=MPI.SUM)
    weighted_sum = totals

    thrust_check_total = 0

    cell_force = np.zeros(len(cell_centre_list_np) * 3)
    thrust_check = cv.convolution_2dkernel_force(
        disc,
        disc_centre,
        disc_radius,
        disc_normal,
        disc_up,
        cell_centre_list_np,
        cell_volume_list_np,
        weighted_sum,
        cell_force,
    )

    thrust_check_array = np.array([thrust_check])
    thrust_check_total_array = np.array([0.0])

    MPI.COMM_WORLD.Allreduce(thrust_check_array, thrust_check_total_array, op=MPI.SUM)
    thrust_check_total = thrust_check_total_array[0]
    thrust_check = thrust_check_total

    # thrust_check = 0.0
    total_thrust = 0.0
    for idx, w in enumerate(weighted_sum):
        segment = disc[idx]
        # if w > 0.0:
        #    thrust_check += segment[0]
        total_thrust += segment[0]

    # Broken: Cell_force_scaled will have a different struct to cell_force
    if thrust_check > 0.0:
        thrust_factor = total_thrust / thrust_check
        # if MPI.COMM_WORLD.Get_rank() == 0:
        cell_force_scaled = []
        for cell in range(len(cell_force) // 3):
            cell_force_scaled.append(
                (
                    cell_force[cell * 3 + 0] * thrust_factor,
                    cell_force[cell * 3 + 1] * thrust_factor,
                    cell_force[cell * 3 + 2] * thrust_factor,
                )
            )
        return cell_force_scaled
    else:
        cell_force = np.ndarray.tolist(cell_force)
        cell_array = iter(cell_force)
        return list(zip(cell_array, cell_array, cell_array))

    cell_force = np.ndarray.tolist(cell_force)
    cell_array = iter(cell_force)
    return list(zip(cell_array, cell_array, cell_array))


def main(
    turbine_zone_dict: dict,
    u_ref: float,
    density: float,
    turbine_name_dict: Optional[dict] = {},
    turbine_name: Optional[str] = "",
    annulusVel: Optional[list] = None,
    annulusTi: Optional[list] = None,
) -> list:
    """
    Creates turbine segments using an actuator disc model.

    Args:
        turbine_zone_dict: Dictionary of turbine zone information.
        v0, v1, v2: Velocity components.
        density: Fluid density.
        turbine_name_dict: (optional) Dictionary of turbine names- required for report output.
        turbine_name: (optional) Name of the turbine- required for report output.
        annulusVel: (optional) Annulus velocity.
        annulusTi: (optional) Annulus turbulence intensity.

    Returns:
        Annulus data list.
    """

    ad = zActuactorDisc(turbine_zone_dict)  # Use more descriptive variable name
    annulus = ad.create_turbine_segments(
        u_ref,
        density,
        annulusVel=annulusVel,
        turbine_name=turbine_name,
        turbine_name_dict=turbine_name_dict,
    )
    return annulus


def create_annulus(turbine_zone_dict: dict) -> list:
    """
    Creates an annulus using an actuator disc model.

    Args:
        turbine_zone_dict: Dictionary of turbine zone information.

    Returns:
        Annulus data list.
    """

    with zActuactorDisc(turbine_zone_dict) as ad:  # Use context manager
        ad.create_annulus()
        annulus = ad.get_annulus()
    return annulus



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Helper functions for plotting zCFD report files
"""

import matplotlib.pyplot as plt
import os
import ipywidgets.widgets as widgets
from IPython.display import display
import pandas as pd
import re
from zutil.fileutils import (
    zCFD_Result,
    get_zcfd_result,
    zCFD_Result,
    zCFD_Overset_Result,
    _get_average_time_from_file,
)
from zutil.analysis.acoustic import calculate_PSD
from zutil.plot import plot_PSD, plot_thirdoctave
from pathlib import Path
from typing import Optional
import warnings
from matplotlib.axes import Axes


class zCFD_Plots:
    """helper class to produce plots for zCFD report files"""

    def __init__(self, control_file_str) -> None:
        self.result = get_zcfd_result(control_file_str)

        if isinstance(self.result, zCFD_Overset_Result):
            self.overset = True
        elif isinstance(self.result, zCFD_Result):
            self.overset = False
        else:
            raise ValueError("Unknown result type")

        self.reports = []

        if self.overset:
            for result in self.result.overset_cases:
                self.reports.append(Report(result.control_file_path))
        else:
            self.reports.append(Report(self.result.control_file_path))

    def plot_residuals(self) -> None:
        """Plot residuals for all reports"""
        for report in self.reports:
            report.plot_residuals()

    def plot_forces(self, variables=None, mean: int = 20) -> None:
        """Plot forces for all reports"""
        for ii, report in enumerate(self.reports):
            report.plot_forces(variables=variables, mean=mean, report_index=ii)

    def plot_performance(self) -> None:
        """Plot performance for all reports"""
        if self.overset:
            # only one log file exists here for overset cases
            print("Log file information is currently limited for overset cases")
            _performance_plot(self.result.log_file_path)
            self.reports[0].plot_linear_solve_memory_usage()
        else:
            for report in self.reports:
                report.plot_performance()

    def get_performance_dict(self, ranks=[0]) -> dict:
        """Get performance data for all reports on the specified ranks"""
        solver_performance = {}
        for report in self.reports:
            report.get_performance_dict(solver_performance, ranks)

        return solver_performance

    def plot_linear_solve_memory_usage(self) -> None:
        """Plot AMGX memory usage for all reports"""
        for report in self.reports:
            report.plot_linear_solve_memory_usage()


class Report(zCFD_Result):
    """Extension of zCFD_Result class to add plotting functions"""

    def __init__(self, control_file: Optional[str] = None) -> None:
        super().__init__(control_file)
        self.cb_container = None
        self.button = None
        self.rolling_avg = 100

        if "BATCH_ANALYSIS" in os.environ:
            self.batch = True
        else:
            self.batch = False

        if not control_file:
            print("Report() is deprecated, use zcfd_plots(path_to_controlfile) instead")
            warnings.warn(
                "Report() is deprecated, use zcfd_plots(path_to_controlfile) instead",
                DeprecationWarning,
            )

    def calculate_PSD(
        self, monitor_name: Optional[str] = None, sampling_frequency=1.0
    ) -> tuple[Optional[list], Optional[list]]:
        """Calculate the Power Spectral Density (PSD) for a given monitor name"""

        if not monitor_name:
            self._print_variables()
            return None, None
        else:
            if monitor_name not in self.report.data.keys():
                raise ValueError(f"Monitor '{monitor_name}' not found in report data.")

            print(f"Calculating PSD for {monitor_name}...")
            print(self.report.data[monitor_name].to_list())
            print(self.parameters["time marching"]["unsteady"]["time step"])
            # Calculate the PSD for the specified monitor
            f, Pxx_den = calculate_PSD(
                self.report.data[monitor_name].to_list(),
                self.parameters["time marching"]["unsteady"]["time step"],
                sampling_frequency=sampling_frequency,
            )

            return f, Pxx_den

    def plot_PSD(
        self, monitor_name: Optional[str] = None, sampling_frequency=1.0
    ) -> Optional[Axes]:
        """Calculate the Power Spectral Density (PSD) for a given monitor name"""

        if not monitor_name:
            self._print_variables()
            return None
        else:
            if monitor_name not in self.report.data.keys():
                raise ValueError(f"Monitor '{monitor_name}' not found in report data.")

            print(f"Calculating PSD for {monitor_name}...")
            print(self.report.data[monitor_name].to_list())
            print(self.parameters["time marching"]["unsteady"]["time step"])
            # Calculate the PSD for the specified monitor
            ax = plot_PSD(
                self.report.data[monitor_name].to_list(),
                self.parameters["time marching"]["unsteady"]["time step"],
                sampling_frequency=sampling_frequency,
            )

            return ax

    def plot_third_octave_bands(
        self, monitor_name: Optional[str] = None, sampling_frequency=1.0
    ) -> Optional[Axes]:
        """Calculate the third-octave bands for a given monitor name"""

        if not monitor_name:
            self._print_variables()
            return None
        else:
            if monitor_name not in self.report.data.keys():
                raise ValueError(f"Monitor '{monitor_name}' not found in report data.")

            print(f"Calculating third-octave bands for {monitor_name}...")
            print(self.report.data[monitor_name].to_list())
            print(self.parameters["time marching"]["unsteady"]["time step"])
            # Calculate the third-octave bands for the specified monitor
            ax = plot_thirdoctave(
                self.report.data[monitor_name].to_list(),
                self.parameters["time marching"]["unsteady"]["time step"],
                sampling_frequency=sampling_frequency,
            )

            return ax

    def plot(self, csv_name):
        """code to support legacy plotting functionality"""
        # remove _report.csv from csv_name
        print("plot(csv_name) function is deprecated, use plot_residuals() instead")
        control_name = csv_name.replace("_report.csv", "")
        self.read_control_file(control_name)
        self.plot_residuals()
        # print depreciation warning

    def plot_residuals(self) -> None:
        """Plot residual data from report_file"""
        fig = plt.figure(figsize=(8, 5), dpi=100)
        ax = fig.gca()

        ax.clear()
        ax.set_yscale("log")

        ax.set_xlabel("cycles")
        ax.set_ylabel("RMS residual")

        ax.set_title(self.control_file_stem)

        # y = self.residual_list
        x = "Cycle"

        for y in self.report.residual_list:
            self.report.data.plot(x=x, y=y, ax=ax, legend=False)

        # Turn on major and minor grid lines
        ax.grid(True, "both")
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        plt.legend(loc="center left", bbox_to_anchor=(1.0, 0.5))

        if self.batch:
            fig.savefig(self.control_file_stem + "_residuals.png", dpi=100)

        plt.show()
        self.visible_fig = []

    def plot_forces(
        self, variables=None, mean: int = 20, report_index: Optional[int] = None
    ) -> None:
        if self.batch:
            if report_index is not None:
                print("report_index: {}".format(report_index))
            print("Plotting forces for case: {}".format(self.control_file_stem))
            self.plot_forces_batch(variables=variables, mean=mean)
        else:
            self.plot_forces_interactive(mean=mean)

    def _print_variables(self):
        """Print available variables to plot"""
        print("Available variables to plot:")
        available_vars = []
        for var in self.report.data.keys():
            if var not in self.report.residual_list + ["RealTimeStep", "Cycle"]:
                available_vars.append(var)

        print(available_vars)

    def plot_forces_batch(self, variables: Optional[list] = None, mean: int = 20):
        if not variables:
            # print available variables
            self._print_variables()

        else:
            # check if variable is valid:
            for var in variables:
                if var not in self.report.data.keys():
                    print(f"ERROR: Variable '{var}' not found in report data.")
                    self._print_variables()
                else:
                    print(f"Plotting '{var}'...  ")
                    rolling = self.report.data.rolling(self.rolling_avg).mean()
                    fig, ax = _forces_plot(self.report, var)
                    _rolling_avg(rolling, var, ax)
                    fig.savefig(f"{self.control_file_stem}_{var}.png")
                    print(f"... Saved {self.control_file_stem}_{var}.png")

    def plot_forces_interactive(self, mean: int = 20) -> None:
        """Create a checkbox widget to plot force and monitor data from a zCFD report file"""
        # Need to disable autoscroll
        # autoscroll(-1)

        self.rolling_avg = mean

        if self.report.header_list is None:
            return

        if self.cb_container is None:
            self.out = widgets.Output()
            self.checkboxes = []
            self.cb_container = widgets.VBox()
            for h in self.report.header_list:
                if h not in self.report.residual_list and h not in [
                    "RealTimeStep",
                    "Cycle",
                ]:
                    self.checkboxes.append(
                        widgets.Checkbox(description=h, value=False, width=90)
                    )

            row_list = []
            for i in range(0, len(self.checkboxes), 3):
                row = widgets.HBox()
                row.children = self.checkboxes[i : i + 3]
                row_list.append(row)

            self.cb_container.children = [i for i in row_list]

            self.rolling = widgets.IntSlider(
                value=self.rolling_avg,
                min=1,
                max=1000,
                step=1,
                description="Rolling Average:",
            )

            self.button = widgets.Button(description="Update plots")
            self.button.on_click(self.plot_data)

        display(self.out)
        display(self.cb_container)
        display(self.rolling)
        display(self.button)

    def plot_data(self, _) -> None:
        """Plot data from checked boxes in the plot_forces widget"""
        self.rolling_avg = self.rolling.value
        self.out.clear_output()
        with self.out:
            rolling = self.report.data.rolling(self.rolling_avg).mean()
            for cb in self.checkboxes:
                if cb.value:
                    y = cb.description
                    fig, ax = _forces_plot(self.report, y)
                    _rolling_avg(rolling, y, ax)
                    self.visible_fig.append(fig)
                    plt.show()

    def plot_performance(self, log_file_path: Optional[str] = None) -> None:
        """Plot CFD solver performance from a zCFD log file"""

        if log_file_path:
            print(
                "plot_performance(log_file_path) function is deprecated, use plot_performance() instead"
            )
            warnings.warn(
                "plot_performance(log_file_path) function is deprecated, use plot_performance() instead",
                DeprecationWarning,
            )

        _performance_plot(self.log_file_path)

        try:
            self.plot_linear_solve_memory_usage()
        except Exception as e:
            print("Error plotting AMGX memory usage: " + str(e))

    def plot_linear_solve_memory_usage(self) -> None:
        """Plots the linear solver usage from all zCFD log file ranks"""
        # figure out the number of ranks
        memory_usage = []

        for rank_log in self.rank_report_paths:
            memory_usage.append(_get_linear_solve_memory_usage_from_file(rank_log))

        fig = plt.figure(figsize=(8, 5), dpi=100)
        ax = fig.gca()
        ax.grid(True)
        ax.set_xlabel("Cycle")
        ax.set_ylabel("AMGX memory usage (MB)")
        ax.set_title("AMGX Memory Usage for case {}".format(self.control_file_stem))

        for ii, rank_memory_usage in enumerate(memory_usage):
            ax.plot(rank_memory_usage)
            if len(rank_memory_usage) > 0:
                ax.plot(rank_memory_usage, label="Rank {}".format(ii))
                if self.batch:
                    fig.savefig(self.control_file_stem + "_memory_usage.png", dpi=100)
                ax.annotate(
                    "Rank {}".format(ii),
                    (len(rank_memory_usage) - 1, rank_memory_usage[-1]),
                    textcoords="offset points",
                    xytext=(20, 0),
                    ha="center",
                )
            else:
                print(
                    "No amgx memory usage data found in log file: {}".format(
                        self.rank_report_paths[ii]
                    )
                )

        ax.set_xlim(ax.get_xlim()[0], ax.get_xlim()[1] * 1.1)

        # ax.legend(loc="upper right")
        plt.show()

    def get_performance_dict(self, performance_dict: dict, ranks=[0]) -> dict:
        """Get performance data from zCFD log file"""

        for rank in ranks:
            rank_log = self.rank_report_paths[rank]
            # get performance data from the log file
            average_output_time = _get_average_time_from_file(rank_log, "outputing")
            average_report_time = _get_average_time_from_file(rank_log, "reporting")
            average_solve_time = _get_average_time_from_file(rank_log, "solving")
            performance_dict[self._control_file_stem]["rank {}".format(rank)] = {
                "average_output_time": average_output_time,
                "average_report_time": average_report_time,
                "average_solve_time": average_solve_time,
            }

        return performance_dict


def _performance_plot(fname: str) -> None:
    if not os.path.isfile(fname):
        print("File not found: " + str(fname))
        return

    elapsed_time = []

    with open(fname) as lfile:
        for line in lfile:
            line = re.findall(r"Timer: Elapsed seconds: [0-9.]+", line)
            if line:
                line = line[0].split(":")[2]
                elapsed_time.append(float(line))

    df = pd.DataFrame(elapsed_time)

    fig = plt.figure(figsize=(8, 5), dpi=100)
    ax = fig.gca()
    ax.grid(True)
    ax.set_xlabel("cycles")
    ax.set_ylabel("Cycle Elapsed time (secs)")
    df.plot(ax=ax, legend=False)
    if "BATCH_ANALYSIS" in os.environ:
        fig.savefig(fname + ".png", dpi=100)
    plt.show()


def _get_linear_solve_memory_usage_from_file(log_file: str) -> list:
    """Extract linear solver memory usage from zCFD log file"""
    if not os.path.isfile(log_file):
        print("File not found: " + str(log_file))
        return []

    memory_usage = []

    with open(log_file, "r") as f:
        lines = f.readlines()
        memory_pattern = re.compile(
            r"Maximum\s+Memory\s+Usage:\s+(\d+(?:\.\d+)?)(\s*(MB|GB|KB))"
        )
        memory_usage = []
        for line in lines:
            match = memory_pattern.search(line)
            if match:
                memory_info = {"value": match.group(1), "unit": match.group(3)}
                memory_usage.append(memory_info)

    main_solve_memory_usage = []

    for ii, memory_info in enumerate(memory_usage):
        # extract every other value
        if ii % 2 == 0:
            value = float(memory_info["value"])
            unit = memory_info["unit"]
            if unit == "KB":
                value *= 1e-3
            elif unit == "GB":
                value *= 1e3
            main_solve_memory_usage.append(value)

    return main_solve_memory_usage


def _forces_plot(report, y_axis):
    fig = plt.figure()
    ax = fig.gca()
    report.data.plot(x="Cycle", y=y_axis, ax=ax, legend=False)
    last_val = report.data[y_axis].tail(1).values

    ax.grid(True)
    ax.set_xlabel("cycles")
    ax.set_ylabel(y_axis)
    ax.set_title(str(y_axis) + " - " + str(last_val))
    return fig, ax


def _rolling_avg(rolling, y_axis, ax):
    rolling.plot(x="Cycle", y=y_axis, ax=ax, legend=False)
    # $\downarrow$ $\uparrow$ $\leftrightarrow$
    rolling_grad = rolling[y_axis].tail(2).values[0] - rolling[y_axis].tail(2).values[1]
    trend = r"$\leftrightarrow$"
    if rolling_grad > 0.0:
        trend = r"$\downarrow$"
    elif rolling_grad < 0.0:
        trend = r"$\uparrow$"
    ax.set_title(ax.get_title() + " " + trend)



"""
Copyright (c) 2012-2024, Zenotech Ltd
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Zenotech Ltd nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ZENOTECH LTD BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

File I/O utilities
"""

import importlib
import os
from pathlib import Path
import json
import glob
import pandas as pd
from typing import Optional, Union
import re
from importlib.resources import files


class zCFD_Report(object):
    """zCFD Base Report object- will handle reading and data IO of report file"""

    def __init__(self, report_file: Optional[str] = None) -> None:
        self.header_list = None

        if report_file:
            if Path(report_file).exists():
                self.read_data(report_file)

    def read_data(self, report_file: str) -> None:
        """Read in csv data from zCFD report file"""
        self.data = get_csv_data(report_file, header=True).dropna(axis=1, how="all")

        # Check for restart by
        restart_file = report_file.rsplit(".csv", 1)[0] + ".restart.csv"
        if os.path.isfile(restart_file):
            self.restart_data = get_csv_data(restart_file, header=True).dropna(
                axis=1, how="all"
            )
            # Get first entry in new data
            restart_cycle = self.data["Cycle"].iloc[0]
            self.restart_data = self.restart_data[
                self.restart_data["Cycle"] < restart_cycle
            ]
            # Merge restart data with data
            self.data = pd.concat([self.restart_data, self.data], ignore_index=True)

        self.header_list: list[str] = list(self.data)
        self.residual_list = []
        for h in self.header_list:
            if h.startswith("rho") or h == "p" or h.startswith("turbEqn"):
                append_str = ""
                # if append_index > 0:
                #    append_str = '_'+str(append_index)
                self.residual_list.append(h + append_str)


class zCFD_Result_Base(object):
    """Base class for zCFD result files"""

    def __init__(self) -> None:
        self.verbose: bool = False
        self.log_file_path: str
        self.report: zCFD_Report

    def _print(self, message):
        if self.verbose:
            print(message)

    def get_average_solve_time(self) -> float:
        """Get average solve time for all reports"""
        return _get_average_time_from_file(self.log_file_path, "solving")

    def get_average_report_time(self) -> float:
        """Get average report time for all reports"""
        return _get_average_time_from_file(self.log_file_path, "reporting")


class zCFD_Result(zCFD_Result_Base):
    """helper class for zCFD result files"""

    def __init__(
        self,
        control_file_path: Optional[str] = None,
        verbose: bool = False,
        absolute=True,
    ) -> None:
        # paths
        self._control_file_path = Path()
        self._report_file_path = Path()
        self._output_directory_path = Path()
        self._volume_file_path = Path()
        self._status_file_path = Path()
        self._control_file_stem = Path()
        self._data_dir = Path()
        self._log_file_path = Path()
        self._checkpoint_path = Path()
        self._mesh_path = Path()

        # boundary file paths- resolve to strings
        self._wall_boundary_path = Path()
        self._symmetry_boundary_path = Path()
        self._farfield_boundary_path = Path()
        self._periodic_boundary_path = Path()
        self._immersed_boundary_path = Path()
        self._inflow_boundary_path = Path()
        self._outflow_boundary_path = Path()
        self._overset_boundary_path = Path()
        self._fwh_wall_data = Path()

        # strings
        # paths
        self.control_file_path: str
        self.report_file_path: str
        self.output_directory_path: str
        self.volume_file_path: str
        self.status_file_path: str
        self.control_file_stem: str
        self.data_dir: str
        self.log_file_path: str
        self.checkpoint_path: str
        self.mesh_path: str

        # boundary file paths- resolve to strings
        self.wall_boundary_path: str
        self.symmetry_boundary_path: str
        self.farfield_boundary_path: str
        self.periodic_boundary_path: str
        self.immersed_boundary_path: str
        self.inflow_boundary_path: str
        self.outflow_boundary_path: str
        self.overset_boundary_path: str
        self.fwh_wall_data: str

        # bools
        self.zcfd_started: bool
        self.overset: bool = False
        self.zcfd_completed: bool
        self.verbose: bool = verbose
        self.steady: bool
        self.absolute = True
        self.rotating = False
        self.translating = False

        # ints
        self.num_procs: int

        # lists
        self.boundary_files: list
        self.immersed_wall_paths: list = []
        self.fwh_permeable_paths: list = []
        self.rank_report_paths: list = []

        # data objects
        self.report: zCFD_Report
        self.parameters: dict
        self.status: dict

        # wizardary to make outside references to path variables resolve to strings

        self._all_paths = [
            attr.lstrip("_")
            for attr, value in vars(self).items()
            if isinstance(value, Path)
        ]

        for path in self._all_paths:
            setattr(
                self.__class__,
                path,
                property(lambda self, p=path: str(getattr(self, f"_{p}"))),
            )

        if control_file_path:
            # check if .py file extension has been removed or not
            self.read_control_file(control_file_path)

    def read_control_file(self, control_file_path_str):
        control_file_path, valid = check_valid_control_file(control_file_path_str)

        if self.absolute:
            control_file_path = control_file_path.absolute()

        if valid:
            self._control_file_path = control_file_path
            self._control_file_stem = Path(control_file_path).stem
            self._data_dir = self._control_file_path.parent

            self.parameters = get_parameters_from_file(control_file_path)

            self._check_steady_state()
            self._check_rotating()
            self._check_translating()

            self._status_file_path = Path(
                self._data_dir, self._control_file_stem + "_status.txt"
            )

            self.zcfd_started = self._check_status()

            if self.zcfd_started:
                self._init_paths()

    def _check_steady_state(self):
        if self.parameters.get("time marching").get("unsteady").get(
            "total time"
        ) == self.parameters.get("time marching").get("unsteady").get("time step"):
            self.steady = True
        else:
            self.steady = False

    def _check_rotating(self):
        """sets true if the fluid domain is rotating"""
        fz_keys = [key for key in self.parameters.keys() if "FZ_" in key]
        if len(fz_keys) == 0:
            self.rotating = False
        else:
            for fz in fz_keys:
                if (self.parameters[fz]["type"] == "rotating") and (
                    0 in self.parameters[fz]["zone"]
                ):
                    self.rotating = True
                    return

    def _check_translating(self):
        """sets true if the fluid domain is translating"""
        fz_keys = [key for key in self.parameters.keys() if "FZ_" in key]
        if len(fz_keys) == 0:
            self.translating = False
        else:
            for fz in fz_keys:
                if (self.parameters[fz]["type"] == "translating") and (
                    0 in self.parameters[fz]["zone"]
                ):
                    self.translating = True
                    return

    def _init_paths(self):
        """initialise paths to the various zCFD files"""
        self._print("initialising paths")

        self._report_file_path = Path(
            self._data_dir, self._control_file_stem + "_report.csv"
        )
        self.report = zCFD_Report(str(self._report_file_path))

        self._status_file_path = Path(
            self._data_dir, self._control_file_stem + "_status.txt"
        )
        self._checkpoint_path = Path(
            self._data_dir, self._control_file_stem + "_results.h5"
        )

        self._mesh_path = Path(self.status["mesh"] + ".h5")

        self.num_procs = self.status["num processor"]

        self._output_directory_path = Path(
            self._data_dir,
            self._control_file_stem + "_P{}_OUTPUT/".format(self.num_procs),
        )

        self._volume_file_path = Path(
            self._output_directory_path, self._control_file_stem + ".pvd"
        )

        self._log_file_path = Path(self._data_dir, self._control_file_stem + ".log")
        self._get_rank_log_paths()

        self._init_boundary_paths()

    def _get_rank_log_paths(self):
        """get paths to rank log files"""
        self._print("getting rank log paths")
        rank_log_dir = Path(self._output_directory_path, "LOGGING")
        for r in range(self.num_procs):
            self.rank_report_paths.append(
                Path(rank_log_dir, self._control_file_stem + "." + str(r) + ".log")
            )

    def _init_boundary_paths(self):
        self._print("getting boundary files")
        self.boundary_files = list(
            self._output_directory_path.glob(self._control_file_stem + "_*.pvd")
        )

        for boundary_file in self.boundary_files:
            # big ugly case switch depending on file extension
            if boundary_file.match("*_symmetry.pvd"):
                self._symmetry_boundary_path = boundary_file
            elif boundary_file.match("*_wall.pvd"):
                self._wall_boundary_path = boundary_file
            elif boundary_file.match("*_immersed wall.pvd"):
                self._immersed_boundary_path = boundary_file
            elif boundary_file.match("*_farfield.pvd"):
                self._farfield_boundary_path = boundary_file
            elif boundary_file.match("*_periodic.pvd"):
                self._periodic_boundary_path = boundary_file
            elif boundary_file.match("*_overset.pvd"):
                self._overset_path = boundary_file
            elif boundary_file.match("*_inflow.pvd"):
                self._inflow_boundary_path = boundary_file
            elif boundary_file.match("*_outflow.pvd"):
                self._outflow_boundary_path = boundary_file
            else:
                print(
                    "error, unrecognised boundary file type: {}".format(boundary_file)
                )

        # if immersed, check for mapped solutions
        if self._immersed_boundary_path:
            immersed_list = self.parameters.get("write output").get(
                "immersed boundary surface"
            )
            if immersed_list:
                for immersed_zone in immersed_list:
                    stl_name = Path(immersed_zone["stl"]).stem
                    # perform regex pattern match of output directory to find all unsteady immersed wall files
                    immersed_files = glob.glob(
                        str(self.output_directory_path) + "/" + stl_name + "_[0-9]*.vtu"
                    )
                    # sort the files by cycle number (glob will put 11 before 9 etc...)
                    immersed_files.sort(key=self._sort_unsteady_key)

                    self.immersed_wall_paths.append(immersed_files)

        # set FWH wall data path
        if self.parameters.get("write output").get("fwh wall data"):
            self._fwh_wall_data = (
                self._output_directory_path
                / "ACOUSTIC_DATA"
                / (str(self._control_file_stem) + "_wall_FWHData.h5")
            )

        # check for permeable FWH surfaces
        if self.parameters.get("write output").get("fwh interpolate"):
            fwh_interpolate_list = self.parameters.get("write output").get(
                "fwh interpolate"
            )
            for fwh_interpolate in fwh_interpolate_list:
                stl_name = Path(fwh_interpolate).stem
                fwh_name = stl_name + "_FWHData.h5"
                self.fwh_permeable_paths.append(
                    self._output_directory_path / "ACOUSTIC_DATA" / fwh_name
                )

    def _check_status(self):
        # Get the status dictionary - this bit is copied from zutil/zutil/post/post.py on Github
        if self._status_file_path.exists():
            with open(self._status_file_path, "r") as f:
                self.status = json.load(f)

            if "problem" in self.status.keys():
                self.status["mesh"] = self.status["problem"]

            return True
        else:
            self._print(
                "Status file path ( {}) does not exist".format(self.status_file_path)
            )
            return False

    @staticmethod
    def _sort_unsteady_key(filename):
        # Extracts the integer (cycle number) from the filename
        match = re.search(r"_(\d+)\.vtu$", filename)
        if match:
            return int(match.group(1))  # Return the integer for sorting
        return 0  # Default return value if the pattern is not found


class zCFD_Overset_Result(zCFD_Result_Base):
    """Class with helper functions for overset result paths"""

    def __init__(
        self,
        override_file_path: Optional[str] = None,
        verbose: bool = False,
        absolute=True,
    ) -> None:
        # paths
        self._override_file_path = Path()
        self._override_file_stem = Path()
        self._log_file_path = Path()
        self._data_dir = Path()

        # strings
        # paths
        self.override_file_path: str
        self.override_file_stem: str
        self.log_file_path: str
        self.data_dir: str

        # bools
        self.verbose: bool = verbose
        self.overset = True
        self.zcfd_started = False

        # ints

        # lists
        self.overset_cases: list[zCFD_Result]

        # data objects
        self.override: dict

        # wizardary to make outside references to path variables resolve to strings

        self._all_paths = [
            attr.lstrip("_")
            for attr, value in vars(self).items()
            if isinstance(value, Path)
        ]

        for path in self._all_paths:
            setattr(
                self.__class__,
                path,
                property(lambda self, p=path: str(getattr(self, f"_{p}"))),
            )

        if override_file_path:
            # check if .py file extension has been removed or not
            self.read_override_file(override_file_path)

    def read_override_file(self, override_file_path) -> None:
        override_file_path, valid = check_valid_control_file(override_file_path)
        if valid:
            self._override_file_path = override_file_path
            self._override_file_stem = Path(override_file_path).stem
            self._data_dir = self._override_file_path.parent

            self.override = get_parameters_from_file(
                override_file_path, parameters_key="override"
            )

            self._log_file_path = Path(self.data_dir, self._override_file_stem + ".log")

            self._init_overset_cases()

    def _init_overset_cases(self) -> None:
        """initialise paths to the various zCFD files"""
        self._print("initialising paths")

        self.overset_cases = []
        for pair in self.override["mesh_case_pair"]:
            control_path = Path(pair[1])
            if control_path.is_absolute():
                overset_case = control_path
            else:
                overset_case = Path(self._data_dir, control_path)

            self.overset_cases.append(
                zCFD_Result(str(overset_case), verbose=self.verbose)
            )
            # recursively set overset flag to True
            self.overset_cases[-1].overset = True

        # get zcfd_started from background case
        self.zcfd_started = self.overset_cases[0].zcfd_started


def get_zcfd_result(
    control_file_path: str, verbose: bool = False, absolute=True
) -> Union[zCFD_Result, zCFD_Overset_Result]:
    """function to read the supplied parameters file, identify if it is an overset case and return the appropriate object"""
    control_file_path, valid = check_valid_control_file(control_file_path)

    # cache cwd
    cwd = os.getcwd()

    # move to file directory for relative imports
    file_dir = Path(control_file_path).parent
    os.chdir(file_dir)
    control_file_path = control_file_path.name

    if valid:
        # check if overset
        namespace = _get_namespace_from_file(control_file_path)
        if namespace.get("override"):
            result = zCFD_Overset_Result(
                control_file_path, verbose=verbose, absolute=absolute
            )
            os.chdir(cwd)
            return result
        elif namespace.get("parameters"):
            result = zCFD_Result(control_file_path, verbose=verbose, absolute=absolute)
            os.chdir(cwd)
            return result
        else:
            raise ValueError(
                "Control file {} does not contain a 'parameters' or 'override' dictionary".format(
                    control_file_path
                )
            )
    else:
        raise FileNotFoundError(
            "Control file {} could not be found".format(control_file_path)
        )


def get_parameters_from_file(
    control_file_path: Path,
    parameters_key: str = "parameters",
    namespace: Optional[dict] = None,
    working_dir: str = ".",
) -> dict:
    """
    Loads and executes a Python module file, extracting the 'parameters' attribute
    as a dictionary.

    Args:
        control_file_path (Path or str): Path to the Python file containing the
                                         `parameters` dictionary.
    Optional Args:
        parameters_key (str): The key to extract from the imported module.
        namespace (dict): The namespace to execute the file in- used to pass variables to the control file.

    Returns:
        dict: The 'parameters' dictionary from the imported module.

    Raises:
        ValueError: If the provided file path does not exist.
        KeyError: If the 'parameters' attribute is not found in the file.
    """
    namespace = namespace or {}

    namespace = _get_namespace_from_file(control_file_path, namespace)

    # check if case is an overset control file
    # Retrieve and return the 'parameters' dictionary from the namespace
    try:
        control_dict = namespace[parameters_key]
    except KeyError:
        raise KeyError(
            "The file does not contain a '{}' attribute".format(parameters_key)
        )

    return control_dict


def _get_namespace_from_file(
    control_file_path: Path, namespace: Optional[dict] = None
) -> dict:
    """
    Loads and executes a Python module file, returning the resulting namespace.

    Args:
        control_file_path (Path): Path to the Python file to be executed"""

    # Get the current __file__ attribute from the namespace - needed to provide __file__ to the control file
    if namespace:
        old_file = namespace.get("__file__")
    else:
        namespace = {}
        old_file = None

    # Convert string input to Path and ensure file has a .py extension
    if isinstance(control_file_path, str):
        if not control_file_path.endswith(".py"):
            control_file_path += ".py"
        control_file_path = Path(control_file_path)
    namespace["__file__"] = str(control_file_path)
    # Check if the file exists before proceeding
    if not control_file_path.exists():
        raise ValueError(
            "Control file path ({}) does not exist".format(control_file_path)
        )

    # Read the file and execute its content in a temporary namespace
    with open(control_file_path, "r") as f:
        control_string = f.read()
    exec(control_string, namespace)

    # reset __file__ attribute to its original value
    if old_file:
        namespace["__file__"] = old_file
    return namespace


def include(filename: str, _globals: Optional[dict] = None) -> None:
    """
    include a file by executing it. This imports everything including
    variables into the calling module
    """
    _globals = _globals or {}
    if os.path.exists(filename):
        exec(compile(open(filename, "rb").read(), filename, "exec"), _globals)


def clean_name(name: str) -> str:
    """Returns the blank name for a zcfd case- removes .py extension if used"""
    return os.path.splitext(name)[0]


def get_zone_info(zone_dict_name: str) -> dict:
    """
    Loads a zone control dictionary by importing it as a Python module.

    Args:
        zone_dict_name (str): The name of the module containing the zone dictionary.

    Returns:
        dict: The loaded zone control dictionary.

    Raises:
        ImportError: If the module cannot be imported or does not contain a dictionary.
        ValueError: If the imported module does not contain a dictionary.
    """

    try:
        zone_module = importlib.import_module(zone_dict_name)
        return zone_module
    except ImportError as e:
        raise ImportError(
            f"Failed to import zone control dictionary from '{zone_dict_name}': {e}"
        ) from e


def check_valid_control_file(control_file_path_str: str) -> tuple[Path, bool]:
    """Util to check the validity of a file path supplied as a zCFD control file"""
    control_file_path = Path(control_file_path_str)
    # check file extension
    if not control_file_path.suffix == ".py":
        # try appending .py first
        control_file_path = control_file_path.with_suffix(".py")
    if not control_file_path.exists():
        print(
            "Error: zCFD control file {} could not be found".format(control_file_path)
        )
        if not control_file_path.is_absolute():
            print("Relative path supplied, cwd is: {}".format(os.getcwd()))
        return control_file_path, False
    return control_file_path, True


def get_csv_data(
    filename: str, header: bool = False, remote: bool = False, delim: str = " "
) -> pd.DataFrame:
    if not header:
        data = pd.read_csv(filename, sep=delim, header=None)
    else:
        data = pd.read_csv(filename, sep=delim)
    return data


def get_fw_csv_data(
    filename: str, widths: int, header: bool = False, remote: bool = False, **kwargs
) -> any:
    """Reads csv and returns specified rows"""
    if not header:
        data = pd.read_fwf(filename, sep=" ", header=None, widths=widths, **kwargs)
    else:
        data = pd.read_fwf(filename, sep=" ", width=widths, **kwargs)
    return data


def _get_logo_path(strapline=False):
    zcfd_home = Path(__file__).parents[5]
    if strapline:
        logo = "ZCFD_Mark_CMYK.png"
    else:
        logo = "ZCFD_Mark_CMYK_No_Strapline_trans.png"
    file_loc = files("zutil").joinpath("assets") / logo
    return str(file_loc)


def _get_average_time_from_file(log_file: str, time_block: str = "solving") -> float:
    """Extract solve time from zCFD log file"""
    if not os.path.isfile(log_file):
        print("File not found: " + str(log_file))
        return 0.0

    average_solve_time = []
    with open(log_file, "r") as f:
        lines = f.readlines()
        time_pattern = re.compile(
            r"{}:\s+\d+\s+days\s+\d+\s+hrs\s+\d+\s+mins\s+(\d+)\s+s\s+\(count:\s+\d+\s+cycles,\s+avg:\s+(\d+(?:\.\d+)?)\s+ms\)".format(
                time_block
            )
        )
        for line in reversed(lines):  # Iterate in reverse to find the last match
            match = time_pattern.search(line)
            if match:
                return float(match.group(2))

    return 0.0  # Default return if no match is found



# kimchi.py
# For converting Python 2 pickles to Python 3

import os
import dill
import pickle


def convert(old_pkl: str) -> str:
    """
    Convert a Python 2 pickle to Python 3

    old_pkl = path to Python 2 pickle file
    """
    # Make a name for the new pickle
    new_pkl = os.path.splitext(old_pkl)[0] + "_p3.pkl"

    print(new_pkl)

    # Convert Python 2 "ObjectType" to Python 3 object
    dill._dill._reverse_typemap["ObjectType"] = object

    # Open the pickle using latin1 encoding
    with open(old_pkl, "rb") as f:
        loaded = pickle.load(f, encoding="latin1")

    # Re-save as Python 3 pickle
    with open(new_pkl, "wb") as outfile:
        pickle.dump(loaded, outfile)
